from __future__ import annotations

import numpy as np

from bec_server.file_writer.default_writer import DefaultFormat


class cSAXSFormat(DefaultFormat):
    """
    cSAXS-specific NeXus format following the NXsas application definition.
    """

    def format(self) -> None:
        """
        Format the data for the cSAXS beamline.
        """

        # local references to keep the code a bit shorter
        storage = self.storage
        data = self.data
        device_manager = self.device_manager
        file_references = self.file_references

        # /entry
        entry = storage.create_group("entry")
        entry.attrs["NX_class"] = "NXentry"
        entry.attrs["definition"] = "NXsas"
        entry.attrs["start_time"] = data.get("start_time")
        entry.attrs["end_time"] = data.get("end_time")
        entry.attrs["version"] = 1.0

        # /entry/collection
        collection = entry.create_group("collection")
        collection.attrs["NX_class"] = "NXcollection"
        bec_collection = collection.create_group("bec")

        # /entry/control
        control = entry.create_group("control")
        control.attrs["NX_class"] = "NXmonitor"
        control.create_dataset(name="mode", data="monitor")
        control.create_dataset(name="integral", data=self.get_entry("bpm4i"))

        # /entry/data
        main_data = entry.create_group("data")
        main_data.attrs["NX_class"] = "NXdata"
        if "eiger_4" in device_manager.devices:
            main_data.create_soft_link(name="data", target="/entry/instrument/eiger_4/data")
        elif "eiger9m" in device_manager.devices:
            main_data.create_soft_link(name="data", target="/entry/instrument/eiger9m/data")
        elif "pilatus_2" in device_manager.devices:
            main_data.create_soft_link(name="data", target="/entry/instrument/pilatus_2/data")

        # /entry/sample
        control = entry.create_group("sample")
        control.attrs["NX_class"] = "NXsample"
        control.create_dataset(name="name", data=self.get_entry("samplename"))
        control.create_dataset(name="description", data=data.get("sample_description"))
        x_translation = control.create_dataset(name="x_translation", data=self.get_entry("samx"))
        x_translation.attrs["units"] = "mm"
        y_translation = control.create_dataset(name="y_translation", data=self.get_entry("samy"))
        y_translation.attrs["units"] = "mm"
        temperature_log = control.create_dataset(
            name="temperature_log", data=self.get_entry("temp")
        )
        temperature_log.attrs["units"] = "K"

        # /entry/instrument
        instrument = entry.create_group("instrument")
        instrument.attrs["NX_class"] = "NXinstrument"
        instrument.create_dataset(name="name", data="cSAXS beamline")

        source = instrument.create_group("source")
        source.attrs["NX_class"] = "NXsource"
        source.create_dataset(name="type", data="Synchrotron X-ray Source")
        source.create_dataset(name="name", data="Swiss Light Source")
        source.create_dataset(name="probe", data="x-ray")
        distance = source.create_dataset(
            name="distance", data=-33800 - np.asarray(self.get_entry("samz", 0))
        )
        distance.attrs["units"] = "mm"
        sigma_x = source.create_dataset(name="sigma_x", data=0.202)
        sigma_x.attrs["units"] = "mm"
        sigma_y = source.create_dataset(name="sigma_y", data=0.018)
        sigma_y.attrs["units"] = "mm"
        divergence_x = source.create_dataset(name="divergence_x", data=0.000135)
        divergence_x.attrs["units"] = "radians"
        divergence_y = source.create_dataset(name="divergence_y", data=0.000025)
        divergence_y.attrs["units"] = "radians"
        current = source.create_dataset(name="current", data=self.get_entry("curr"))
        current.attrs["units"] = "mA"

        insertion_device = instrument.create_group("insertion_device")
        insertion_device.attrs["NX_class"] = "NXinsertion_device"
        source.create_dataset(name="type", data="undulator")
        gap = source.create_dataset(name="gap", data=self.get_entry("idgap"))
        gap.attrs["units"] = "mm"
        k = source.create_dataset(name="k", data=2.46)
        k.attrs["units"] = "NX_DIMENSIONLESS"
        length = source.create_dataset(name="length", data=1820)
        length.attrs["units"] = "mm"

        slit_0 = instrument.create_group("slit_0")
        slit_0.attrs["NX_class"] = "NXslit"
        source.create_dataset(name="material", data="OFHC Cu")
        source.create_dataset(name="description", data="Horizontal secondary source slit")
        x_gap = source.create_dataset(name="x_gap", data=self.get_entry("sl0wh"))
        x_gap.attrs["units"] = "mm"
        x_translation = source.create_dataset(name="x_translation", data=self.get_entry("sl0ch"))
        x_translation.attrs["units"] = "mm"
        distance = source.create_dataset(
            name="distance", data=-21700 - np.asarray(self.get_entry("samz", 0))
        )
        distance.attrs["units"] = "mm"

        slit_1 = instrument.create_group("slit_1")
        slit_1.attrs["NX_class"] = "NXslit"
        source.create_dataset(name="material", data="OFHC Cu")
        source.create_dataset(name="description", data="Horizontal secondary source slit")
        x_gap = source.create_dataset(name="x_gap", data=self.get_entry("sl1wh"))
        x_gap.attrs["units"] = "mm"
        y_gap = source.create_dataset(name="y_gap", data=self.get_entry("sl1wv"))
        y_gap.attrs["units"] = "mm"
        x_translation = source.create_dataset(name="x_translation", data=self.get_entry("sl1ch"))
        x_translation.attrs["units"] = "mm"
        height = source.create_dataset(name="x_translation", data=self.get_entry("sl1ch"))
        height.attrs["units"] = "mm"
        distance = source.create_dataset(
            name="distance", data=-7800 - np.asarray(self.get_entry("samz", 0))
        )
        distance.attrs["units"] = "mm"

        mono = instrument.create_group("monochromator")
        mono.attrs["NX_class"] = "NXmonochromator"
        mokev = data.get("mokev", {})
        if mokev:
            if isinstance(mokev, list):
                mokev = mokev[0]
            wavelength = mono.create_dataset(
                name="wavelength", data=12.3984193 / (mokev.get("mokev").get("value") + 1e-9)
            )
            wavelength.attrs["units"] = "Angstrom"
            energy = mono.create_dataset(name="energy", data=mokev.get("mokev").get("value"))
            energy.attrs["units"] = "keV"
        mono.create_dataset(name="type", data="Double crystal fixed exit monochromator.")
        distance = mono.create_dataset(
            name="distance", data=-5220 - np.asarray(self.get_entry("samz", 0))
        )
        distance.attrs["units"] = "mm"

        crystal_1 = mono.create_group("crystal_1")
        crystal_1.attrs["NX_class"] = "NXcrystal"
        crystal_1.create_dataset(name="usage", data="Bragg")
        crystal_1.create_dataset(name="order_no", data="1")
        crystal_1.create_dataset(name="reflection", data="[1 1 1]")
        bragg_angle = crystal_1.create_dataset(name="bragg_angle", data=self.get_entry("moth1"))
        bragg_angle.attrs["units"] = "degrees"

        crystal_2 = mono.create_group("crystal_2")
        crystal_2.attrs["NX_class"] = "NXcrystal"
        crystal_2.create_dataset(name="usage", data="Bragg")
        crystal_2.create_dataset(name="order_no", data="2")
        crystal_2.create_dataset(name="reflection", data="[1 1 1]")
        bragg_angle = crystal_2.create_dataset(name="bragg_angle", data=self.get_entry("moth1"))
        bragg_angle.attrs["units"] = "degrees"
        bend_x = crystal_2.create_dataset(name="bend_x", data=self.get_entry("mobd"))
        bend_x.attrs["units"] = "degrees"

        xbpm4 = instrument.create_group("XBPM4")
        xbpm4.attrs["NX_class"] = "NXdetector"
        xbpm4_sum = xbpm4.create_group("XBPM4_sum")
        xbpm4_sum_data = xbpm4_sum.create_dataset(name="data", data=self.get_entry("bpm4s"))
        xbpm4_sum_data.attrs["units"] = "NX_DIMENSIONLESS"
        xbpm4_sum.create_dataset(name="description", data="Sum of counts for the four quadrants.")
        xbpm4_x = xbpm4.create_group("XBPM4_x")
        xbpm4_x_data = xbpm4_x.create_dataset(name="data", data=self.get_entry("bpm4x"))
        xbpm4_x_data.attrs["units"] = "NX_DIMENSIONLESS"
        xbpm4_x.create_dataset(
            name="description",
            data="Normalized difference of counts between left and right quadrants.",
        )
        xbpm4_y = xbpm4.create_group("XBPM4_y")
        xbpm4_y_data = xbpm4_y.create_dataset(name="data", data=self.get_entry("bpm4y"))
        xbpm4_y_data.attrs["units"] = "NX_DIMENSIONLESS"
        xbpm4_y.create_dataset(
            name="description",
            data="Normalized difference of counts between high and low quadrants.",
        )
        xbpm4_skew = xbpm4.create_group("XBPM4_skew")
        xbpm4_skew_data = xbpm4_skew.create_dataset(name="data", data=self.get_entry("bpm4z"))
        xbpm4_skew_data.attrs["units"] = "NX_DIMENSIONLESS"
        xbpm4_skew.create_dataset(
            name="description", data="Normalized difference of counts between diagonal quadrants."
        )

        mirror = instrument.create_group("mirror")
        mirror.attrs["NX_class"] = "NXmirror"
        mirror.create_dataset(name="type", data="single")
        mirror.create_dataset(
            name="description",
            data="Grazing incidence mirror to reject high-harmonic wavelengths from the monochromator. There are three coating options available that are used depending on the X-ray energy, no coating (SiO2), rhodium (Rh) or platinum (Pt).",
        )
        incident_angle = mirror.create_dataset(name="incident_angle", data=self.get_entry("mith"))
        incident_angle.attrs["units"] = "degrees"
        substrate_material = mirror.create_dataset(name="substrate_material", data="SiO2")
        substrate_material.attrs["units"] = "NX_CHAR"
        coating_material = mirror.create_dataset(name="coating_material", data="SiO2")
        coating_material.attrs["units"] = "NX_CHAR"
        bend_y = mirror.create_dataset(name="bend_y", data="mibd")
        bend_y.attrs["units"] = "NX_DIMENSIONLESS"
        distance = mirror.create_dataset(
            name="distance", data=-4370 - np.asarray(self.get_entry("samz", 0))
        )
        distance.attrs["units"] = "mm"

        xbpm5 = instrument.create_group("XBPM5")
        xbpm5.attrs["NX_class"] = "NXdetector"
        xbpm5_sum = xbpm5.create_group("XBPM5_sum")
        xbpm5_sum_data = xbpm5_sum.create_dataset(name="data", data=self.get_entry("bpm5s"))
        xbpm5_sum_data.attrs["units"] = "NX_DIMENSIONLESS"
        xbpm5_sum.create_dataset(name="description", data="Sum of counts for the four quadrants.")
        xbpm5_x = xbpm5.create_group("XBPM5_x")
        xbpm5_x_data = xbpm5_x.create_dataset(name="data", data=self.get_entry("bpm5x"))
        xbpm5_x_data.attrs["units"] = "NX_DIMENSIONLESS"
        xbpm5_x.create_dataset(
            name="description",
            data="Normalized difference of counts between left and right quadrants.",
        )
        xbpm5_y = xbpm5.create_group("XBPM5_y")
        xbpm5_y_data = xbpm5_y.create_dataset(name="data", data=self.get_entry("bpm5y"))
        xbpm5_y_data.attrs["units"] = "NX_DIMENSIONLESS"
        xbpm5_y.create_dataset(
            name="description",
            data="Normalized difference of counts between high and low quadrants.",
        )
        xbpm5_skew = xbpm5.create_group("XBPM5_skew")
        xbpm5_skew_data = xbpm5_skew.create_dataset(name="data", data=self.get_entry("bpm5z"))
        xbpm5_skew_data.attrs["units"] = "NX_DIMENSIONLESS"
        xbpm5_skew.create_dataset(
            name="description", data="Normalized difference of counts between diagonal quadrants."
        )

        slit_2 = instrument.create_group("slit_2")
        slit_2.attrs["NX_class"] = "NXslit"
        source.create_dataset(name="material", data="Ag")
        source.create_dataset(name="description", data="Slit 2, optics hutch")
        x_gap = source.create_dataset(name="x_gap", data=self.get_entry("sl2wh"))
        x_gap.attrs["units"] = "mm"
        y_gap = source.create_dataset(name="y_gap", data=self.get_entry("sl2wv"))
        y_gap.attrs["units"] = "mm"
        x_translation = source.create_dataset(name="x_translation", data=self.get_entry("sl2ch"))
        x_translation.attrs["units"] = "mm"
        height = source.create_dataset(name="x_translation", data=self.get_entry("sl2cv"))
        height.attrs["units"] = "mm"
        distance = source.create_dataset(
            name="distance", data=-3140 - np.asarray(self.get_entry("samz", 0))
        )
        distance.attrs["units"] = "mm"

        slit_3 = instrument.create_group("slit_3")
        slit_3.attrs["NX_class"] = "NXslit"
        source.create_dataset(name="material", data="Si")
        source.create_dataset(name="description", data="Slit 3, experimental hutch, exposure box")
        x_gap = source.create_dataset(name="x_gap", data=self.get_entry("sl3wh"))
        x_gap.attrs["units"] = "mm"
        y_gap = source.create_dataset(name="y_gap", data=self.get_entry("sl3wv"))
        y_gap.attrs["units"] = "mm"
        x_translation = source.create_dataset(name="x_translation", data=self.get_entry("sl3ch"))
        x_translation.attrs["units"] = "mm"
        height = source.create_dataset(name="x_translation", data=self.get_entry("sl3cv"))
        height.attrs["units"] = "mm"
        # distance = source.create_dataset(name="distance", data=-3140 - self.get_entry("samz", 0))
        # distance.attrs["units"] = "mm"

        filter_set = instrument.create_group("filter_set")
        filter_set.attrs["NX_class"] = "NXattenuator"
        filter_set.create_dataset(name="material", data="Si")
        filter_set.create_dataset(
            name="description",
            data="The filter set consists of 4 linear stages, each with five filter positions. Additionally, each one allows for an out position to allow 'no filtering'.",
        )
        attenuator_transmission = filter_set.create_dataset(
            name="attenuator_transmission", data=10 ** self.get_entry("ftrans", 0)
        )
        attenuator_transmission.attrs["units"] = "NX_DIMENSIONLESS"

        slit_4 = instrument.create_group("slit_4")
        slit_4.attrs["NX_class"] = "NXslit"
        source.create_dataset(name="material", data="Si")
        source.create_dataset(name="description", data="Slit 4, experimental hutch, exposure box")
        x_gap = source.create_dataset(name="x_gap", data=self.get_entry("sl4wh"))
        x_gap.attrs["units"] = "mm"
        y_gap = source.create_dataset(name="y_gap", data=self.get_entry("sl4wv"))
        y_gap.attrs["units"] = "mm"
        x_translation = source.create_dataset(name="x_translation", data=self.get_entry("sl4ch"))
        x_translation.attrs["units"] = "mm"
        height = source.create_dataset(name="x_translation", data=self.get_entry("sl4cv"))
        height.attrs["units"] = "mm"
        # distance = source.create_dataset(name="distance", data=-3140 - self.get_entry("samz", 0))
        # distance.attrs["units"] = "mm"

        slit_5 = instrument.create_group("slit_5")
        slit_5.attrs["NX_class"] = "NXslit"
        source.create_dataset(name="material", data="Si")
        source.create_dataset(name="description", data="Slit 5, experimental hutch, exposure box")
        x_gap = source.create_dataset(name="x_gap", data=self.get_entry("sl5wh"))
        x_gap.attrs["units"] = "mm"
        y_gap = source.create_dataset(name="y_gap", data=self.get_entry("sl5wv"))
        y_gap.attrs["units"] = "mm"
        x_translation = source.create_dataset(name="x_translation", data=self.get_entry("sl5ch"))
        x_translation.attrs["units"] = "mm"
        height = source.create_dataset(name="x_translation", data=self.get_entry("sl5cv"))
        height.attrs["units"] = "mm"
        # distance = source.create_dataset(name="distance", data=-3140 - self.get_entry("samz", 0))
        # distance.attrs["units"] = "mm"

        beam_stop_1 = instrument.create_group("beam_stop_1")
        beam_stop_1.attrs["NX_class"] = "NX_beamstop"
        beam_stop_1.create_dataset(name="description", data="circular")
        bms1_size = beam_stop_1.create_dataset(name="size", data=3)
        bms1_size.attrs["units"] = "mm"
        bms1_x = beam_stop_1.create_dataset(name="size", data=self.get_entry("bs1x"))
        bms1_x.attrs["units"] = "mm"
        bms1_y = beam_stop_1.create_dataset(name="size", data=self.get_entry("bs1y"))
        bms1_y.attrs["units"] = "mm"

        beam_stop_2 = instrument.create_group("beam_stop_2")
        beam_stop_2.attrs["NX_class"] = "NX_beamstop"
        beam_stop_2.create_dataset(name="description", data="rectangular")
        bms2_size_x = beam_stop_2.create_dataset(name="size_x", data=5)
        bms2_size_x.attrs["units"] = "mm"
        bms2_size_y = beam_stop_2.create_dataset(name="size_y", data=2.25)
        bms2_size_y.attrs["units"] = "mm"
        bms2_x = beam_stop_2.create_dataset(name="size", data=self.get_entry("bs2x"))
        bms2_x.attrs["units"] = "mm"
        bms2_y = beam_stop_2.create_dataset(name="size", data=self.get_entry("bs2y"))
        bms2_y.attrs["units"] = "mm"
        bms2_data = beam_stop_2.create_dataset(name="data", data=self.get_entry("diode"))
        bms2_data.attrs["units"] = "NX_DIMENSIONLESS"

        if "eiger1p5m" in device_manager.devices and device_manager.devices.eiger1p5m.enabled:
            eiger_4 = instrument.create_group("eiger_4")
            eiger_4.attrs["NX_class"] = "NXdetector"
            x_pixel_size = eiger_4.create_dataset(name="x_pixel_size", data=75)
            x_pixel_size.attrs["units"] = "um"
            y_pixel_size = eiger_4.create_dataset(name="y_pixel_size", data=75)
            y_pixel_size.attrs["units"] = "um"
            polar_angle = eiger_4.create_dataset(name="polar_angle", data=0)
            polar_angle.attrs["units"] = "degrees"
            azimuthal_angle = eiger_4.create_dataset(name="azimuthal_angle", data=0)
            azimuthal_angle.attrs["units"] = "degrees"
            rotation_angle = eiger_4.create_dataset(name="rotation_angle", data=0)
            rotation_angle.attrs["units"] = "degrees"
            description = eiger_4.create_dataset(
                name="description", data="Single-photon counting detector, 320 micron-thick Si chip"
            )
            orientation = eiger_4.create_group("orientation")
            orientation.attrs["description"] = (
                "Orientation defines the number of counterclockwise rotations by 90 deg followed by a transposition to reach the 'cameraman orientation', that is looking towards the beam."
            )
            orientation.create_dataset(name="transpose", data=1)
            orientation.create_dataset(name="rot90", data=3)

        if (
            "eiger9m" in device_manager.devices
            and device_manager.devices.eiger9m.enabled
            and "eiger9m" in file_references
        ):
            eiger9m = instrument.create_group("eiger9m")
            eiger9m.attrs["NX_class"] = "NXdetector"
            x_pixel_size = eiger9m.create_dataset(name="x_pixel_size", data=75)
            x_pixel_size.attrs["units"] = "um"
            y_pixel_size = eiger9m.create_dataset(name="y_pixel_size", data=75)
            y_pixel_size.attrs["units"] = "um"
            polar_angle = eiger9m.create_dataset(name="polar_angle", data=0)
            polar_angle.attrs["units"] = "degrees"
            azimuthal_angle = eiger9m.create_dataset(name="azimuthal_angle", data=0)
            azimuthal_angle.attrs["units"] = "degrees"
            rotation_angle = eiger9m.create_dataset(name="rotation_angle", data=0)
            rotation_angle.attrs["units"] = "degrees"
            description = eiger9m.create_dataset(
                name="description",
                data="Eiger9M detector, in-house developed, Paul Scherrer Institute",
            )
            orientation = eiger9m.create_group("orientation")
            orientation.attrs["description"] = (
                "Orientation defines the number of counterclockwise rotations by 90 deg followed by a transposition to reach the 'cameraman orientation', that is looking towards the beam."
            )
            orientation.create_dataset(name="transpose", data=1)
            orientation.create_dataset(name="rot90", data=3)
            data = eiger9m.create_ext_link("data", file_references["eiger9m"]["path"], "EG9M/data")
            status = eiger9m.create_ext_link(
                "status", file_references["eiger9m"]["path"], "EG9M/status"
            )

        if (
            "pilatus_2" in device_manager.devices
            and device_manager.devices.pilatus_2.enabled
            and "pilatus_2" in file_references
        ):
            pilatus_2 = instrument.create_group("pilatus_2")
            pilatus_2.attrs["NX_class"] = "NXdetector"
            x_pixel_size = pilatus_2.create_dataset(name="x_pixel_size", data=172)
            x_pixel_size.attrs["units"] = "um"
            y_pixel_size = pilatus_2.create_dataset(name="y_pixel_size", data=172)
            y_pixel_size.attrs["units"] = "um"
            polar_angle = pilatus_2.create_dataset(name="polar_angle", data=0)
            polar_angle.attrs["units"] = "degrees"
            azimuthal_angle = pilatus_2.create_dataset(name="azimuthal_angle", data=0)
            azimuthal_angle.attrs["units"] = "degrees"
            rotation_angle = pilatus_2.create_dataset(name="rotation_angle", data=0)
            rotation_angle.attrs["units"] = "degrees"
            description = pilatus_2.create_dataset(
                name="description", data="Pilatus 300K detector, Dectris, Switzerland"
            )
            orientation = pilatus_2.create_group("orientation")
            orientation.attrs["description"] = (
                "Orientation defines the number of counterclockwise rotations by 90 deg followed by a transposition to reach the 'cameraman orientation', that is looking towards the beam."
            )
            orientation.create_dataset(name="transpose", data=1)
            orientation.create_dataset(name="rot90", data=2)
            data = pilatus_2.create_ext_link(
                "data", file_references["pilatus_2"]["path"], "entry/instrument/pilatus_2/data"
            )

        if (
            "falcon" in device_manager.devices
            and device_manager.devices.falcon.enabled
            and "falcon" in file_references
        ):
            falcon = instrument.create_ext_link(
                "falcon", file_references["falcon"]["path"], "entry/instrument/FalconX1"
            )



from unittest import mock

import pytest

from bec_lib.device import Device, DeviceBase, Positioner
from bec_server.scan_server.scan_manager import ScanManager
from bec_server.scan_server.scans import ScanArgType


@pytest.fixture
def scan_manager():
    parent = mock.MagicMock()
    yield ScanManager(parent=parent)


@pytest.mark.parametrize(
    "arg_input, arg_output",
    [
        ({"a": float}, {"a": "float"}),
        ({"a": ScanArgType.FLOAT}, {"a": "float"}),
        ({"a": ScanArgType.DEVICE}, {"a": "device"}),
        ({"a": ScanArgType.INT}, {"a": "int"}),
        ({"a": ScanArgType.BOOL}, {"a": "boolean"}),
        ({"a": ScanArgType.LIST}, {"a": "list"}),
        ({"a": ScanArgType.DICT}, {"a": "dict"}),
        ({"a": str}, {"a": "str"}),
        ({"a": int}, {"a": "int"}),
        ({"a": bool}, {"a": "boolean"}),
        ({"a": list}, {"a": "list"}),
        ({"a": dict}, {"a": "dict"}),
        ({"a": DeviceBase}, {"a": "device"}),
        ({"a": Device}, {"a": "device"}),
        ({"a": Positioner}, {"a": "device"}),
    ],
)
def test_scan_manager_convert_arg_input(scan_manager, arg_input, arg_output):
    assert scan_manager.convert_arg_input(arg_input) == arg_output



import pytest
from pydantic import ValidationError

from bec_server.scan_server.scan_gui_models import GUIConfig
from bec_server.scan_server.scans import ScanArgType, ScanBase


class GoodScan(ScanBase):  # pragma: no cover
    scan_name = "good_scan"
    required_kwargs = ["steps", "relative"]
    arg_input = {
        "device": ScanArgType.DEVICE,
        "start": ScanArgType.FLOAT,
        "stop": ScanArgType.FLOAT,
    }
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": None}
    gui_config = {"Scan Parameters": ["steps", "exp_time", "relative", "burst_at_each_point"]}

    def __init__(
        self,
        *args,
        exp_time: float = 0,
        steps: int = None,
        relative: bool = False,
        burst_at_each_point: int = 1,
        **kwargs,
    ):
        """
        A good scan for one or more motors.

        Args:
            *args (Device, float, float): pairs of device / start position / end position
            exp_time (float): exposure time in s. Default: 0
            steps (int): number of steps. Default: 10
            relative (bool): if True, the start and end positions are relative to the current position. Default: False
            burst_at_each_point (int): number of acquisition per point. Default: 1

        Returns:
            ScanReport


        """
        super().__init__(
            exp_time=exp_time, relative=relative, burst_at_each_point=burst_at_each_point, **kwargs
        )
        self.steps = steps

    def doing_something_good(self):
        pass


class ExtraKwarg(ScanBase):  # pragma: no cover
    scan_name = "wrong_name"
    required_kwargs = ["steps", "relative"]

    gui_config = {"Device 1": ["motor1", "start_motor1", "stop_motor1"]}

    def __init__(self, motor1: str, stop_motor1: float, **kwargs):
        """
        A scan following Fermat's spiral.

        Args:
            motor1 (DeviceBase): first motor
            start_motor1 (float): start position motor 1
            stop_motor1 (float): end position motor 1

        Returns:
            ScanReport
        """

    print("I am a wrong scan")


class WrongDocs(ScanBase):  # pragma: no cover
    scan_name = "wrong_name"
    required_kwargs = ["steps", "relative"]

    gui_config = {"Device 1": ["motor1", "start_motor1", "stop_motor1"]}

    def __init__(self, motor1: str, start_motor1: float, stop_motor1: float, **kwargs):
        """
        A scan following Fermat's spiral.

        Args:
            motor1 (DeviceBase): first motor

        Returns:
            ScanReport
        """

    print("I am a scan with wrong docs.")


def test_gui_config_good_scan_dump():
    gui_config = GUIConfig.from_dict(GoodScan)
    expected_config = {
        "scan_class_name": "GoodScan",
        "arg_group": {
            "name": "Scan Arguments",
            "bundle": 3,
            "arg_inputs": {
                "device": ScanArgType.DEVICE,
                "start": ScanArgType.FLOAT,
                "stop": ScanArgType.FLOAT,
            },
            "inputs": [
                {
                    "arg": True,
                    "name": "device",
                    "display_name": "Device",
                    "type": "device",
                    "tooltip": None,
                    "default": None,
                    "expert": False,
                },
                {
                    "arg": True,
                    "name": "start",
                    "display_name": "Start",
                    "type": "float",
                    "tooltip": None,
                    "default": None,
                    "expert": False,
                },
                {
                    "arg": True,
                    "name": "stop",
                    "display_name": "Stop",
                    "type": "float",
                    "tooltip": None,
                    "default": None,
                    "expert": False,
                },
            ],
            "min": 1,
            "max": None,
        },
        "kwarg_groups": [
            {
                "name": "Scan Parameters",
                "inputs": [
                    {
                        "arg": False,
                        "name": "steps",
                        "display_name": "Steps",
                        "type": "int",
                        "tooltip": "Number of steps",
                        "default": None,
                        "expert": False,
                    },
                    {
                        "arg": False,
                        "name": "exp_time",
                        "display_name": "Exp Time",
                        "type": "float",
                        "tooltip": "Exposure time in s",
                        "default": 0,
                        "expert": False,
                    },
                    {
                        "arg": False,
                        "name": "relative",
                        "display_name": "Relative",
                        "type": "bool",
                        "tooltip": "If True, the start and end positions are relative to the current position",
                        "default": False,
                        "expert": False,
                    },
                    {
                        "arg": False,
                        "name": "burst_at_each_point",
                        "display_name": "Burst At Each Point",
                        "type": "int",
                        "tooltip": "Number of acquisition per point",
                        "default": 1,
                        "expert": False,
                    },
                ],
            }
        ],
    }
    assert gui_config.model_dump() == expected_config


def test_gui_config_extra_kwarg():
    with pytest.raises(ValidationError) as excinfo:
        GUIConfig.from_dict(ExtraKwarg)
    errors = excinfo.value.errors()
    assert len(errors) == 5
    assert errors[0]["type"] == ("wrong argument name")
    assert errors[1]["type"] == ("missing argument name")
    assert errors[2]["type"] == ("missing argument name")
    assert errors[3]["type"] == ("missing argument name")
    assert errors[4]["type"] == ("missing argument name")


def test_gui_config_wrong_docs():
    gui_config = GUIConfig.from_dict(WrongDocs)
    expected = {
        "scan_class_name": "WrongDocs",
        "arg_group": {
            "name": "Scan Arguments",
            "bundle": 0,
            "arg_inputs": {},
            "inputs": [],
            "min": None,
            "max": None,
        },
        "kwarg_groups": [
            {
                "name": "Device 1",
                "inputs": [
                    {
                        "arg": False,
                        "name": "motor1",
                        "type": "str",
                        "display_name": "Motor 1",
                        "tooltip": "First motor",
                        "default": "_empty",
                        "expert": False,
                    },
                    {
                        "arg": False,
                        "name": "start_motor1",
                        "type": "float",
                        "display_name": "Start Motor 1",
                        "tooltip": None,
                        "default": "_empty",
                        "expert": False,
                    },
                    {
                        "arg": False,
                        "name": "stop_motor1",
                        "type": "float",
                        "display_name": "Stop Motor 1",
                        "tooltip": None,
                        "default": "_empty",
                        "expert": False,
                    },
                ],
            }
        ],
    }
    assert gui_config.model_dump() == expected



from unittest import mock

from bec_server.data_processing.cli.launch import main


def test_main():
    with mock.patch(
        "bec_server.data_processing.cli.launch.parse_cmdline_args", return_value=(None, None, None)
    ) as mock_parser:
        with mock.patch("bec_server.data_processing.dap_server.DAPServer") as mock_data_processing:
            with mock.patch("bec_server.data_processing.cli.launch.threading.Event") as mock_event:
                main()
                mock_parser.assert_called_once()
                mock_data_processing.assert_called_once()
                mock_event.assert_called_once()


def test_main_shutdown():
    with mock.patch(
        "bec_server.data_processing.cli.launch.parse_cmdline_args", return_value=(None, None, None)
    ) as mock_parser:
        with mock.patch("bec_server.data_processing.dap_server.DAPServer") as mock_data_processing:
            with mock.patch("bec_server.data_processing.cli.launch.threading.Event") as mock_event:
                mock_event.return_value.wait.side_effect = KeyboardInterrupt()
                main()
                mock_parser.assert_called_once()
                mock_data_processing.assert_called_once()
                mock_event.assert_called_once()
                mock_data_processing.return_value.shutdown.assert_called_once()



"""
Scan Manager loads the available scans and publishes them to redis.
"""

import inspect

from bec_lib import plugin_helper
from bec_lib.device import DeviceBase
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.messages import AvailableResourceMessage
from bec_lib.signature_serializer import signature_to_dict
from bec_server.scan_server.scan_gui_models import GUIConfig

from . import scans as ScanServerScans

logger = bec_logger.logger


class ScanManager:
    """
    Scan Manager loads the available scans and publishes them to redis.
    """

    def __init__(self, *, parent):
        """
        Scan Manager loads and manages the available scans.
        """
        self.parent = parent
        self.available_scans = {}
        self.scan_dict = {}
        self._plugins = {}
        self.load_plugins()
        self.update_available_scans()
        self.publish_available_scans()

    def load_plugins(self):
        """load scan plugins"""
        plugins = plugin_helper.get_scan_plugins()
        if not plugins:
            return
        for name, cls in plugins.items():
            if not issubclass(cls, ScanServerScans.RequestBase):
                logger.error(
                    f"Plugin {name} is not a valid scan plugin as it does not inherit from RequestBase. Skipping."
                )
                continue
            self._plugins[name] = cls
            logger.info(f"Loading scan plugin {name}")

    def update_available_scans(self):
        """load all scans and plugin scans"""
        members = inspect.getmembers(ScanServerScans)
        for member_name, cls in self._plugins.items():
            members.append((member_name, cls))

        for name, scan_cls in members:
            try:
                is_scan = issubclass(scan_cls, ScanServerScans.RequestBase)
            except TypeError:
                is_scan = False

            if not is_scan or not scan_cls.scan_name:
                logger.debug(f"Ignoring {name}")
                continue
            if scan_cls.scan_name in self.available_scans:
                logger.error(f"{scan_cls.scan_name} already exists. Skipping.")
                continue

            report_classes = [
                ScanServerScans.RequestBase,
                ScanServerScans.ScanBase,
                ScanServerScans.AsyncFlyScanBase,
                ScanServerScans.SyncFlyScanBase,
                ScanServerScans.ScanStubs,
                ScanServerScans.ScanComponent,
            ]

            for report_cls in report_classes:
                if issubclass(scan_cls, report_cls):
                    base_cls = report_cls.__name__
            self.scan_dict[scan_cls.__name__] = scan_cls
            gui_config = self.validate_gui_config(scan_cls)
            self.available_scans[scan_cls.scan_name] = {
                "class": scan_cls.__name__,
                "base_class": base_cls,
                "arg_input": self.convert_arg_input(scan_cls.arg_input),
                "gui_config": gui_config,
                "required_kwargs": scan_cls.required_kwargs,
                "arg_bundle_size": scan_cls.arg_bundle_size,
                "doc": scan_cls.__doc__ or scan_cls.__init__.__doc__,
                "signature": signature_to_dict(scan_cls.__init__),
            }

    def validate_gui_config(self, scan_cls) -> dict:
        """
        Validate the gui_config of the scan class

        Args:
            scan_cls: class

        Returns:
            dict: gui_config
        """

        if not hasattr(scan_cls, "gui_config"):
            return {}
        if not isinstance(scan_cls.gui_config, GUIConfig) and not isinstance(
            scan_cls.gui_config, dict
        ):
            logger.error(
                f"Invalid gui_config for {scan_cls.scan_name}. gui_config must be of type GUIConfig or dict."
            )
            return {}
        gui_config = scan_cls.gui_config
        if isinstance(scan_cls.gui_config, dict):
            gui_config = GUIConfig.from_dict(scan_cls)
        return gui_config.model_dump()

    def convert_arg_input(self, arg_input) -> dict:
        """
        Convert the arg_input to supported data types

        Args:
            arg_input: dict

        Returns:
            dict: converted arg_input
        """
        for key, value in arg_input.items():
            if isinstance(value, ScanServerScans.ScanArgType):
                continue
            if issubclass(value, DeviceBase):
                # once we have generalized the device types, this should be removed
                arg_input[key] = "device"
            elif issubclass(value, bool):
                # should be unified with the ScanArgType.BOOL
                arg_input[key] = "boolean"
            else:
                arg_input[key] = value.__name__
        return arg_input

    def publish_available_scans(self):
        """send all available scans to the broker"""
        self.parent.connector.set(
            MessageEndpoints.available_scans(),
            AvailableResourceMessage(resource=self.available_scans),
        )



"""This module contains the base class for SLS detectors. We follow the approach to integrate
PSI detectors into the BEC system based on this base class. The base class is used to implement
certain methods that are expected by BEC, such as stage, unstage, trigger, stop, etc...
We use composition with a custom prepare class to implement BL specific logic for the detector.
The beamlines need to inherit from the CustomDetectorMixing for their mixin classes."""

import os
import threading
import time
import traceback

from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.file_utils import FileWriter
from bec_lib.logger import bec_logger
from ophyd import Component, Device, DeviceStatus, Kind
from ophyd.device import Staged

from ophyd_devices.sim.sim_signals import SetableSignal
from ophyd_devices.utils import bec_utils
from ophyd_devices.utils.bec_scaninfo_mixin import BecScaninfoMixin
from ophyd_devices.utils.errors import DeviceStopError, DeviceTimeoutError

logger = bec_logger.logger


class DetectorInitError(Exception):
    """Raised when initiation of the device class fails,
    due to missing device manager or not started in sim_mode."""


class CustomDetectorMixin:
    """
    Mixin class for custom detector logic

    This class is used to implement BL specific logic for the detector.
    It is used in the PSIDetectorBase class.

    For the integration of a new detector, the following functions should
    help with integrating functionality, but additional ones can be added.

    Check PSIDetectorBase for the functions that are called during relevant function calls of
    stage, unstage, trigger, stop and _init.
    """

    def __init__(self, *_args, parent: Device = None, **_kwargs) -> None:
        self.parent = parent

    def on_init(self) -> None:
        """
        Init sequence for the detector
        """

    def on_stage(self) -> None:
        """
        Specify actions to be executed during stage in preparation for a scan.
        self.parent.scaninfo already has all current parameters for the upcoming scan.

        In case the backend service is writing data on disk, this step should include publishing
        a file_event and file_message to BEC to inform the system where the data is written to.

        IMPORTANT:
        It must be safe to assume that the device is ready for the scan
        to start immediately once this function is finished.
        """

    def on_unstage(self) -> None:
        """
        Specify actions to be executed during unstage.

        This step should include checking if the acqusition was successful,
        and publishing the file location and file event message,
        with flagged done to BEC.
        """

    def on_stop(self) -> None:
        """
        Specify actions to be executed during stop.
        This must also set self.parent.stopped to True.

        This step should include stopping the detector and backend service.
        """

    def on_trigger(self) -> None | DeviceStatus:
        """
        Specify actions to be executed upon receiving trigger signal.
        Return a DeviceStatus object or None
        """

    def on_pre_scan(self) -> None:
        """
        Specify actions to be executed right before a scan starts.

        Only use if needed, and it is recommended to keep this function as short/fast as possible.
        """

    def on_complete(self) -> None | DeviceStatus:
        """
        Specify actions to be executed when the scan is complete.

        This can for instance be to check with the detector and backend if all data is written succsessfully.
        """

    def publish_file_location(self, done: bool, successful: bool, metadata: dict = None) -> None:
        """
        Publish the filepath to REDIS.

        We publish two events here:
        - file_event: event for the filewriter
        - public_file: event for any secondary service (e.g. radial integ code)

        Args:
            done (bool): True if scan is finished
            successful (bool): True if scan was successful
            metadata (dict): additional metadata to publish
        """
        if metadata is None:
            metadata = {}

        msg = messages.FileMessage(
            file_path=self.parent.filepath.get(),
            done=done,
            successful=successful,
            metadata=metadata,
        )
        pipe = self.parent.connector.pipeline()
        self.parent.connector.set_and_publish(
            MessageEndpoints.public_file(self.parent.scaninfo.scan_id, self.parent.name),
            msg,
            pipe=pipe,
        )
        self.parent.connector.set_and_publish(
            MessageEndpoints.file_event(self.parent.name), msg, pipe=pipe
        )
        pipe.execute()

    def wait_for_signals(
        self,
        signal_conditions: list[tuple],
        timeout: float,
        check_stopped: bool = False,
        interval: float = 0.05,
        all_signals: bool = False,
    ) -> bool:
        """
        Convenience wrapper to allow waiting for signals to reach a certain condition.
        For EPICs PVs, an example usage is pasted at the bottom.

        Args:
            signal_conditions (list[tuple]): tuple of executable calls for conditions (get_current_state, condition) to check
            timeout (float): timeout in seconds
            interval (float): interval in seconds
            all_signals (bool): True if all signals should be True, False if any signal should be True

        Returns:
            bool: True if all signals are in the desired state, False if timeout is reached

            >>> Example usage for EPICS PVs:
            >>> self.wait_for_signals(signal_conditions=[(self.acquiring.get, False)], timeout=5, interval=0.05, check_stopped=True, all_signals=True)
        """

        timer = 0
        while True:
            checks = [
                get_current_state() == condition
                for get_current_state, condition in signal_conditions
            ]
            if check_stopped is True and self.parent.stopped is True:
                return False
            if (all_signals and all(checks)) or (not all_signals and any(checks)):
                return True
            if timer > timeout:
                return False
            time.sleep(interval)
            timer += interval

    def wait_with_status(
        self,
        signal_conditions: list[tuple],
        timeout: float,
        check_stopped: bool = False,
        interval: float = 0.05,
        all_signals: bool = False,
        exception_on_timeout: Exception = None,
    ) -> DeviceStatus:
        """Utility function to wait for signals in a thread.
        Returns a DevicesStatus object that resolves either to set_finished or set_exception.
        The DeviceStatus is attached to the parent device, i.e. the detector object inheriting from PSIDetectorBase.

        Usage:
        This function should be used to wait for signals to reach a certain condition, especially in the context of
        on_trigger and on_complete. If it is not used, functions may block and slow down the performance of BEC.
        It will return a DeviceStatus object that is to be returned from the function. Once the conditions are met,
        the DeviceStatus will be set to set_finished in case of success or set_exception in case of a timeout or exception.
        The exception can be specified with the exception_on_timeout argument. The default exception is a TimeoutError.

        Args:
            signal_conditions (list[tuple]): tuple of executable calls for conditions (get_current_state, condition) to check
            timeout (float): timeout in seconds
            check_stopped (bool): True if stopped flag should be checked
            interval (float): interval in seconds
            all_signals (bool): True if all signals should be True, False if any signal should be True
            exception_on_timeout (Exception): Exception to raise on timeout

        Returns:
            DeviceStatus: DeviceStatus object that resolves either to set_finished or set_exception
        """
        if exception_on_timeout is None:
            exception_on_timeout = DeviceTimeoutError(
                f"Timeout error for {self.parent.name} while waiting for signals {signal_conditions}"
            )

        status = DeviceStatus(self.parent)

        # utility function to wrap the wait_for_signals function
        def wait_for_signals_wrapper(
            status: DeviceStatus,
            signal_conditions: list[tuple],
            timeout: float,
            check_stopped: bool,
            interval: float,
            all_signals: bool,
            exception_on_timeout: Exception,
        ):
            """Convenient wrapper around wait_for_signals to set status based on the result.

            Args:
                status (DeviceStatus): DeviceStatus object to be set
                signal_conditions (list[tuple]): tuple of executable calls for conditions (get_current_state, condition) to check
                timeout (float): timeout in seconds
                check_stopped (bool): True if stopped flag should be checked
                interval (float): interval in seconds
                all_signals (bool): True if all signals should be True, False if any signal should be True
                exception_on_timeout (Exception): Exception to raise on timeout
            """
            try:
                result = self.wait_for_signals(
                    signal_conditions, timeout, check_stopped, interval, all_signals
                )
                if result:
                    status.set_finished()
                else:
                    if self.parent.stopped:
                        # INFO This will execute a callback to the parent device.stop() method
                        status.set_exception(exc=DeviceStopError(f"{self.parent.name} was stopped"))
                    else:
                        # INFO This will execute a callback to the parent device.stop() method
                        status.set_exception(exc=exception_on_timeout)
            # pylint: disable=broad-except
            except Exception as exc:
                content = traceback.format_exc()
                logger.warning(
                    f"Error in wait_for_signals in {self.parent.name}; Traceback: {content}"
                )
                # INFO This will execute a callback to the parent device.stop() method
                status.set_exception(exc=exc)

        thread = threading.Thread(
            target=wait_for_signals_wrapper,
            args=(
                status,
                signal_conditions,
                timeout,
                check_stopped,
                interval,
                all_signals,
                exception_on_timeout,
            ),
            daemon=True,
        )
        thread.start()
        return status


class PSIDetectorBase(Device):
    """
    Abstract base class for SLS detectors

    Class attributes:
        custom_prepare_cls (object): class for custom prepare logic (BL specific)

    Args:
        prefix (str): EPICS PV prefix for component (optional)
        name (str): name of the device, as will be reported via read()
        kind (str): member of class 'ophydobj.Kind', defaults to Kind.normal
                    omitted -> readout ignored for read 'ophydobj.read()'
                    normal -> readout for read
                    config -> config parameter for 'ophydobj.read_configuration()'
                    hinted -> which attribute is readout for read
        parent (object): instance of the parent device
        device_manager (object): bec device manager
        **kwargs: keyword arguments
    """

    filepath = Component(SetableSignal, value="", kind=Kind.config)

    custom_prepare_cls = CustomDetectorMixin

    def __init__(self, prefix="", *, name, kind=None, parent=None, device_manager=None, **kwargs):
        super().__init__(prefix=prefix, name=name, kind=kind, parent=parent, **kwargs)
        self.stopped = False
        self.name = name
        self.service_cfg = None
        self.scaninfo = None
        self.filewriter = None

        if not issubclass(self.custom_prepare_cls, CustomDetectorMixin):
            raise DetectorInitError("Custom prepare class must be subclass of CustomDetectorMixin")
        self.custom_prepare = self.custom_prepare_cls(parent=self, **kwargs)

        if device_manager:
            self._update_service_config()
            self.device_manager = device_manager
        else:
            self.device_manager = bec_utils.DMMock()
            base_path = kwargs["basepath"] if "basepath" in kwargs else "."
            self.service_cfg = {"base_path": os.path.abspath(base_path)}

        self.connector = self.device_manager.connector
        self._update_scaninfo()
        self._update_filewriter()
        self._init()

    def _update_filewriter(self) -> None:
        """Update filewriter with service config"""
        self.filewriter = FileWriter(service_config=self.service_cfg, connector=self.connector)

    def _update_scaninfo(self) -> None:
        """Update scaninfo from BecScaninfoMixing
        This depends on device manager and operation/sim_mode
        """
        self.scaninfo = BecScaninfoMixin(self.device_manager)
        self.scaninfo.load_scan_metadata()

    def _update_service_config(self) -> None:
        """Update service config from BEC service config

        If bec services are not running and SERVICE_CONFIG is NONE, we fall back to the current directory.
        """
        # pylint: disable=import-outside-toplevel
        from bec_lib.bec_service import SERVICE_CONFIG

        if SERVICE_CONFIG:
            self.service_cfg = SERVICE_CONFIG.config["service_config"]["file_writer"]
            return
        self.service_cfg = {"base_path": os.path.abspath(".")}

    def check_scan_id(self) -> None:
        """Checks if scan_id has changed and set stopped flagged to True if it has."""
        old_scan_id = self.scaninfo.scan_id
        self.scaninfo.load_scan_metadata()
        if self.scaninfo.scan_id != old_scan_id:
            self.stopped = True

    def _init(self) -> None:
        """Initialize detector, filewriter and set default parameters"""
        self.custom_prepare.on_init()

    def stage(self) -> list[object]:
        """
        Stage device in preparation for a scan.
        First we check if the device is already staged. Stage is idempotent,
        if staged twice it should raise (we let ophyd.Device handle the raise here).
        We reset the stopped flag and get the scaninfo from BEC, before calling custom_prepare.on_stage.

        Returns:
            list(object): list of objects that were staged

        """
        if self._staged != Staged.no:
            return super().stage()
        self.stopped = False
        self.scaninfo.load_scan_metadata()
        self.custom_prepare.on_stage()
        return super().stage()

    def pre_scan(self) -> None:
        """Pre-scan logic.

        This function will be called from BEC directly before the scan core starts, and should only implement
        time-critical actions. Therefore, it should also be kept as short/fast as possible.
        I.e. Arming a detector in case there is a risk of timing out.
        """
        self.custom_prepare.on_pre_scan()

    def trigger(self) -> DeviceStatus:
        """Trigger the detector, called from BEC."""
        # pylint: disable=assignment-from-no-return
        status = self.custom_prepare.on_trigger()
        if isinstance(status, DeviceStatus):
            return status
        return super().trigger()

    def complete(self) -> None:
        """Complete the acquisition, called from BEC.

        This function is called after the scan is complete, just before unstage.
        We can check here with the data backend and detector if the acquisition successfully finished.

        Actions are implemented in custom_prepare.on_complete since they are beamline specific.
        """
        # pylint: disable=assignment-from-no-return
        status = self.custom_prepare.on_complete()
        if isinstance(status, DeviceStatus):
            return status
        status = DeviceStatus(self)
        status.set_finished()
        return status

    def unstage(self) -> list[object]:
        """
        Unstage device after a scan.

        We first check if the scanID has changed, thus, the scan was unexpectedly interrupted but the device was not stopped.
        If that is the case, the stopped flag is set to True, which will immediately unstage the device.

        Custom_prepare.on_unstage is called to allow for BL specific logic to be executed.

        Returns:
            list(object): list of objects that were unstaged
        """
        self.check_scan_id()
        self.custom_prepare.on_unstage()
        self.stopped = False
        return super().unstage()

    def stop(self, *, success=False) -> None:
        """
        Stop the scan, with camera and file writer

        """
        self.custom_prepare.on_stop()
        super().stop(success=success)
        self.stopped = True



# This file is the entry point for the Scan Server.
# It is called either by the bec-scan-server entry point or directly from the command line.

import threading

from bec_lib.bec_service import parse_cmdline_args
from bec_lib.logger import bec_logger
from bec_lib.redis_connector import RedisConnector
from bec_server import scan_server

logger = bec_logger.logger
bec_logger.level = bec_logger.LOGLEVEL.INFO


def main():
    """
    Launch the scan server.
    """
    _, _, config = parse_cmdline_args()

    bec_server = scan_server.scan_server.ScanServer(config=config, connector_cls=RedisConnector)
    try:
        event = threading.Event()
        # pylint: disable=E1102
        logger.success("Started ScanServer")
        event.wait()
    except KeyboardInterrupt:
        bec_server.shutdown()
        event.set()


if __name__ == "__main__":
    main()



from bec_lib import messages


class EmitterBase:
    def __init__(self, connector) -> None:
        self.connector = connector

    def on_init(self, scan_id: str):
        pass

    def on_scan_point_emit(self, scan_id: str, point_id: int):
        pass

    def on_baseline_emit(self, scan_id: str):
        pass

    def on_cleanup(self, scan_id: str):
        pass

    def on_scan_status_update(self, status_msg: messages.ScanStatusMessage):
        pass

    def shutdown(self):
        pass



"""
This module contains the BECLogger class, which is a wrapper around the loguru logger. It is used to
configure and manage the logging of the BEC.
"""

from __future__ import annotations

import enum
import json
import os
import sys
import traceback
from itertools import takewhile
from typing import TYPE_CHECKING

# TODO: Importing bec_lib, instead of `from bec_lib.messages import LogMessage`, avoids potential
# logger <-> messages circular import. But there could be a better solution.
import bec_lib
from bec_lib.bec_errors import ServiceConfigError
from bec_lib.endpoints import MessageEndpoints
from bec_lib.utils.import_utils import lazy_import_from

if TYPE_CHECKING:
    from bec_lib.connector import ConnectorBase

loguru_logger = lazy_import_from("loguru", ("logger",))
LogWriter = lazy_import_from("bec_lib.file_utils", ("LogWriter",))


class LogLevel(int, enum.Enum):
    """Mapping of Loguru log levels to BEC log levels."""

    TRACE = 5
    DEBUG = 10
    INFO = 20
    SUCCESS = 25
    WARNING = 30
    ERROR = 40
    CRITICAL = 50
    CONSOLE_LOG = 21


class BECLogger:
    """Logger for BEC."""

    LOG_FORMAT_STDERR = (
        "<green>{service_name} | {{time:YYYY-MM-DD HH:mm:ss}}</green> | {{name}} | <level>[{{level}}]</level> |"
        " <level>{{message}}</level>\n"
    )
    LOG_FORMAT = (
        "<green>{{time:YYYY-MM-DD HH:mm:ss}}</green> | {{name}} | <level>[{{level}}]</level> |"
        " <level>{{message}}</level>\n"
    )
    DEBUG_FORMAT = (
        "<green>{service_name} | {{time:YYYY-MM-DD HH:mm:ss.SSS}}</green> | <level>{{level}}</level> |"
        "  <level>{{thread.name}} ({{thread.id}})</level> | <cyan>{{name}}</cyan>:<cyan>{{function}}</cyan>:<cyan>{{line}}</cyan> -"
        " <level>{{message}}</level>\n"
    )
    TRACE_FORMAT = (
        "<green>{service_name} | {{time:YYYY-MM-DD HH:mm:ss.SSS}}</green> | <level>{{level}}</level> |"
        " <level>{{thread.name}} ({{thread.id}})</level> | <cyan>{{extra[stack]}}</cyan> - <level>{{message}}</level>\n"
    )
    LOGLEVEL = LogLevel

    _logger = None

    def __init__(self) -> None:
        if hasattr(self, "_configured"):
            return
        self.bootstrap_server = None
        self.connector = None
        self.service_name = None
        self.writer_mixin = None
        self._base_path = None
        self.logger = loguru_logger
        self._log_level = LogLevel.INFO
        self._redis_log_level = self._log_level
        self._stderr_log_level = self._log_level
        self._file_log_level = self._log_level
        self._console_log = False
        self._configured = False
        self._disabled_modules = set()

    def __new__(cls):
        if not hasattr(cls, "_logger") or cls._logger is None:
            cls._logger = super(BECLogger, cls).__new__(cls)
        return cls._logger

    @classmethod
    def _reset_singleton(cls):
        if cls._logger is not None:
            cls._logger.logger.remove()
        cls._logger = None

    def configure(
        self,
        bootstrap_server: list,
        connector_cls: ConnectorBase,
        service_name: str,
        service_config: dict = None,
    ) -> None:
        """
        Configure the logger.

        Args:
            bootstrap_server (list): List of bootstrap servers.
            connector_cls (ConnectorBase): Connector class.
            service_name (str): Name of the service to which the logger belongs.
        """
        if self._configured:
            # already configured, nothing to do - this can happen
            # if running another BECClient (or BECService) in addition
            # to a main one
            return
        if not self._base_path:
            self._update_base_path(service_config)
        if os.path.exists(self._base_path) is False:
            self.writer_mixin.create_directory(self._base_path)
        self.bootstrap_server = bootstrap_server
        self.connector = connector_cls(bootstrap_server)
        self.service_name = service_name
        self._configured = True
        self._update_sinks()

    def _update_base_path(self, service_config: dict = None):
        """
        Compile the log base path.
        """
        # pylint: disable=import-outside-toplevel
        if service_config:
            service_cfg = service_config.get("log_writer", None)
            if not service_cfg:
                raise ServiceConfigError(
                    f"ServiceConfig {service_config} must at least contain key with 'log_writer'"
                )
        else:
            service_cfg = {"base_path": "./"}
        self.writer_mixin = LogWriter(service_cfg)
        self._base_path = self.writer_mixin.directory
        self.writer_mixin.create_directory(self._base_path)

    def _logger_callback(self, msg):
        if not self._configured:
            return
        msg = json.loads(msg)
        msg["service_name"] = self.service_name
        try:
            self.connector.send(
                topic=MessageEndpoints.log(),
                msg=bec_lib.messages.LogMessage(
                    log_type=msg["record"]["level"]["name"].lower(), log_msg=msg
                ),
            )
        except Exception:
            # connector disconnected?
            # just ignore the error here...
            # Exception is not explicitely specified,
            # because it depends on the connector
            pass

    def get_format(self, level: LogLevel = None, is_stderr=False) -> str:
        """
        Get the format for a specific log level.

        Args:
            level (LogLevel, optional): Log level. Defaults to None. If None, the current log level will be used.
            is_stderr (bool, optional): Whether the log is for stderr. Defaults to False.

        Returns:
            str: Log format.
        """
        service_name = self.service_name if self.service_name else ""
        if level is None:
            level = self.level
        if level > self.LOGLEVEL.DEBUG:
            if is_stderr:
                return self.LOG_FORMAT_STDERR.format(service_name=service_name)
            return self.LOG_FORMAT.format(service_name=service_name)
        if level > self.LOGLEVEL.TRACE:
            return self.DEBUG_FORMAT.format(service_name=service_name)
        return self.TRACE_FORMAT.format(service_name=service_name)

    def formatting(self, is_stderr=False):
        """
        Format the log message.

        Args:
            record (dict): Log record.

        Returns:
            dict: Formatted log record.
        """

        def _update_record(record):
            level = record["level"].no
            if level <= self.LOGLEVEL.TRACE:
                frames = takewhile(
                    lambda f: "/loguru/" not in f.filename, traceback.extract_stack()
                )
                stack = " > ".join("{}:{}:{}".format(f.filename, f.name, f.lineno) for f in frames)
                record["extra"]["stack"] = stack
            return level

        def _format(record):
            level = _update_record(record)
            return self.get_format(level)

        def _format_stderr(record):
            level = _update_record(record)
            return self.get_format(level, is_stderr=True)

        if is_stderr:
            return _format_stderr
        return _format

    def _update_sinks(self):
        self.logger.remove()
        self.add_redis_log(self._redis_log_level)
        self.add_sys_stderr(self._stderr_log_level)
        self.add_file_log(self._file_log_level)
        if self._console_log:
            self.add_console_log()

    def filter(self, is_console: bool = False):
        """
        Filter factory function for log messages.

        Args:
            is_console (bool, optional): Whether the log is for the console. Defaults to False.

        Returns:
            function: Filter function.
        """

        def _filter(record):
            if record["name"] in self._disabled_modules:
                return False
            for module in self._disabled_modules:
                if record["name"].startswith(module):
                    return False
            if not is_console and record["level"].no == LogLevel.CONSOLE_LOG:
                return False
            return True

        return _filter

    def add_sys_stderr(self, level: LogLevel):
        """
        Add a sink to stderr.

        Args:
            level (LogLevel): Log level.
        """
        self.logger.add(
            sys.__stderr__,
            level=level,
            format=self.formatting(is_stderr=True),
            filter=self.filter(),
        )

    def add_file_log(self, level: LogLevel):
        """
        Add a sink to the service log file.

        Args:
            level (LogLevel): Log level.
        """
        if not self.service_name:
            return
        filename = os.path.join(self._base_path, f"{self.service_name}.log")
        self.logger.add(
            filename,
            level=level,
            format=self.formatting(),
            filter=self.filter(),
            retention="3 days",
            rotation="3 days",
        )

    def add_console_log(self):
        """
        Add a sink to the console log.
        """
        try:
            self.logger.level("CONSOLE_LOG", no=21, color="<yellow>", icon="📣")
        except TypeError:
            # level with same severity already exists: already configured
            pass

        if not self.service_name:
            return
        filename = os.path.join(self._base_path, f"{self.service_name}_CONSOLE.log")

        # define a level corresponding to console log - this is to be able to filter messages
        # (only those with this particular level will be recorded by the console logger,
        # while other loggers will ignore them)

        self.logger.add(
            filename,
            level=LogLevel.CONSOLE_LOG,
            format=self.get_format(LogLevel.CONSOLE_LOG).rstrip(),
            filter=self.filter(is_console=True),
            retention="3 days",
            rotation="3 days",
        )
        self._console_log = True

    def add_redis_log(self, level: LogLevel):
        """
        Add a sink to the redis log.

        Args:
            level (LogLevel): Log level.
        """
        self.logger.add(
            self._logger_callback,
            serialize=True,
            level=level,
            format=self.formatting(),
            filter=self.filter(),
        )

    @property
    def disabled_modules(self) -> set[str]:
        """
        Get the disabled modules.
        """
        return self._disabled_modules

    @disabled_modules.setter
    def disabled_modules(self, module_names: str | list[str]) -> None:
        """
        Disable log messages from specific modules.

        Args:
            module_names (str | list[str]): Module name(s).
        """
        if isinstance(module_names, str):
            module_names = [module_names]
        self._disabled_modules.update(module_names)

    @property
    def level(self):
        """
        Get the current log level.
        """
        return self._log_level

    @level.setter
    def level(self, val: LogLevel):
        self._log_level = val
        self._redis_log_level = val
        self._file_log_level = val
        self._stderr_log_level = val
        self._update_sinks()


bec_logger = BECLogger()



from __future__ import annotations

import re
from contextvars import ContextVar
from typing import Any, Literal, Optional, Type

from pydantic import BaseModel, Field, field_validator
from pydantic_core import PydanticCustomError

from bec_lib.signature_serializer import signature_to_dict
from bec_server.scan_server.scans import ScanBase

context_signature = ContextVar("context_signature")
context_docstring = ContextVar("context_docstring")


class GUIInput(BaseModel):
    """
    InputConfig is a data model for the input configuration of a scan.

    Args:
        name (str): name of the input, has to be same as in the function signature.
    """

    arg: bool = Field(False)
    name: str = Field(None, validate_default=True)
    type: Optional[
        Literal["DeviceBase", "device", "float", "int", "bool", "str", "list", "dict"]
    ] = Field(None, validate_default=True)
    display_name: Optional[str] = Field(None, validate_default=True)
    tooltip: Optional[str] = Field(None, validate_default=True)
    default: Optional[Any] = Field(None, validate_default=True)
    expert: Optional[bool] = Field(False)  # TODO decide later how to implement

    @field_validator("name")
    @classmethod
    def validate_name(cls, v, values):
        # args cannot be validated with the current implementation of signature of scans
        if values.data["arg"]:
            return v
        signature = context_signature.get()
        available_args = [entry["name"] for entry in signature]
        if v not in available_args:
            raise PydanticCustomError(
                "wrong argument name",
                "The argument name is not available in the function signature",
                {"wrong_value": v},
            )
        return v

    @field_validator("type")
    @classmethod
    def validate_field(cls, v, values):
        # args cannot be validated with the current implementation of signature of scans
        if values.data["arg"]:
            return v
        signature = context_signature.get()
        if v is None:
            name = values.data.get("name", None)
            if name is None:
                raise PydanticCustomError(
                    "missing argument name",
                    "The argument name is required to infer the type",
                    {"wrong_value": v},
                )
            for entry in signature:
                if entry["name"] == name:
                    v = entry["annotation"]
                    return v

    @field_validator("tooltip")
    @classmethod
    def validate_tooltip(cls, v, values):
        # args cannot be validated with the current implementation of signature of scans
        if values.data["arg"]:
            return v
        if v is not None:
            return v

        docstring = context_docstring.get()
        name = values.data.get("name", None)
        if name is None:
            raise PydanticCustomError(
                "missing argument name",
                "The argument name is required to infer the tooltip",
                {"wrong_value": v},
            )

        try:
            args_part = docstring.split("Args:")[1].split("Returns:")[0]
        except IndexError:
            return None

        pattern = re.compile(r"\s*" + re.escape(name) + r" \(([^)]+)\): (.+?)(?:\.|\n|$)")
        match = pattern.search(args_part)

        if match:
            description = match.group(2)
            first_sentence = description.split(".")[0].strip()
            if first_sentence:
                v = first_sentence[0].upper() + first_sentence[1:]
                return v
        return None

    @field_validator("display_name")
    @classmethod
    def validate_display_name(cls, v, values):
        if v is not None:
            return v
        name = values.data.get("name", None)
        if name is None:
            raise PydanticCustomError(
                "missing argument name",
                "The argument name is required to infer the display name",
                {"wrong_value": v},
            )
        parts = re.split("(_|\d+)", name)
        formatted_parts = []
        for part in parts:
            if part.isdigit():
                formatted_parts.append("" + part)
            elif part.isalpha():
                formatted_parts.append(part.capitalize())
        v = " ".join(formatted_parts).strip()
        return v

    @field_validator("default")
    @classmethod
    def validate_default(cls, v, values):
        # args cannot be validated with the current implementation of signature of scans
        if values.data["arg"]:
            return v
        if v is not None:
            return v
        signature = context_signature.get()
        name = values.data.get("name", None)
        if name is None:
            raise PydanticCustomError(
                "missing argument name",
                "The argument name is required to infer the type",
                {"wrong_value": v},
            )
        for entry in signature:
            if entry["name"] == name:
                v = entry["default"]
                return v


class GUIGroup(BaseModel):
    """
    GUIGroup is a data model for the GUI group configuration of a scan.
    """

    name: str
    inputs: list[GUIInput]


class GUIArgGroup(BaseModel):
    """
    GUIArgGroup is a data model for the GUI group configuration of a scan.
    """

    name: str = "Scan Arguments"
    bundle: int = Field(None)
    arg_inputs: dict
    inputs: Optional[list[GUIInput]] = Field(None, validate_default=True)
    min: Optional[int] = Field(None)
    max: Optional[int] = Field(None)

    @field_validator("inputs")
    @classmethod
    def validate_inputs(cls, v, values):
        if v is not None:
            return v
        arg_inputs = values.data["arg_inputs"]
        arg_inputs_str = {key: value.value for key, value in arg_inputs.items()}
        v = []
        for name, type_ in arg_inputs_str.items():
            v.append(GUIInput(name=name, type=type_, arg=True))
        return v


class GUIConfig(BaseModel):
    """
    GUIConfig is a data model for the GUI configuration of a scan.
    """

    scan_class_name: str
    arg_group: Optional[GUIArgGroup] = Field(None)
    kwarg_groups: list[GUIGroup] = Field(None)
    signature: list[dict] = Field(..., exclude=True)
    docstring: str = Field(..., exclude=True)

    @classmethod
    def from_dict(cls, scan_cls: Type[ScanBase]) -> GUIConfig:
        """
        Create a GUIConfig object from a scan class.

        Args:
            scan_cls(Type[ScanBase]): scan class

        Returns:
            GUIConfig: GUIConfig object
        """

        groups = []
        config = scan_cls.gui_config
        signature = signature_to_dict(scan_cls.__init__)
        signature_token = context_signature.set(signature)
        docstring = scan_cls.__doc__ or scan_cls.__init__.__doc__
        docstring_token = context_docstring.set(docstring)
        # kwargs from gui config
        for group_name, input_names in config.items():
            inputs = [GUIInput(name=name, arg=False) for name in input_names]
            group = GUIGroup(name=group_name, inputs=inputs)
            groups.append(group)
        # args from arg_input if available
        arg_group = None
        if hasattr(scan_cls, "arg_input"):
            arg_input = scan_cls.arg_input
            if hasattr(scan_cls, "arg_bundle_size"):
                arg_group = GUIArgGroup(
                    bundle=scan_cls.arg_bundle_size.get("bundle"),
                    min=scan_cls.arg_bundle_size.get("min"),
                    max=scan_cls.arg_bundle_size.get("max"),
                    arg_inputs=arg_input,
                )
            else:
                arg_group = GUIArgGroup(inputs=scan_cls.arg_input)

        context_signature.reset(signature_token)
        context_docstring.reset(docstring_token)
        return cls(
            scan_class_name=scan_cls.__name__,
            signature=signature,
            docstring=docstring,
            kwarg_groups=groups,
            arg_group=arg_group,
        )



# Description: Launch the SciHub connector.
# This script is the entry point for the SciHub connector. It is called either
# by the bec-dap entry point or directly from the command line.
import argparse
import threading

from bec_lib.bec_service import parse_cmdline_args
from bec_lib.logger import bec_logger
from bec_lib.redis_connector import RedisConnector
from bec_server import scihub

logger = bec_logger.logger
bec_logger.level = bec_logger.LOGLEVEL.INFO


def main():
    """
    Launch the SciHub connector.
    """
    _, _, config = parse_cmdline_args()

    sh = scihub.SciHub(config, RedisConnector)

    try:
        event = threading.Event()
        logger.success("Started SciHub connector")
        event.wait()
    except KeyboardInterrupt:
        sh.shutdown()


if __name__ == "__main__":
    main()



import traceback
from threading import Thread

import numpy as np
from bec_lib.logger import bec_logger
from ophyd import Component as Cpt
from ophyd import DeviceStatus, Kind

from ophyd_devices.interfaces.base_classes.psi_detector_base import (
    CustomDetectorMixin,
    PSIDetectorBase,
)
from ophyd_devices.sim.sim_data import SimulatedDataCamera
from ophyd_devices.sim.sim_signals import ReadOnlySignal, SetableSignal
from ophyd_devices.sim.sim_utils import H5Writer
from ophyd_devices.utils.errors import DeviceStopError

logger = bec_logger.logger


class SimCameraSetup(CustomDetectorMixin):
    """Mixin class for the SimCamera device."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._thread_trigger = None
        self._thread_complete = None

    def on_trigger(self) -> None:
        """Trigger the camera to acquire images.

        This method can be called from BEC during a scan. It will acquire images and send them to BEC.
        Whether the trigger is send from BEC is determined by the softwareTrigger argument in the device config.

        Here, we also run a callback on SUB_MONITOR to send the image data the device_monitor endpoint in BEC.
        """
        status = DeviceStatus(self.parent)

        def on_trigger_call(status: DeviceStatus) -> None:
            try:
                for _ in range(self.parent.burst.get()):
                    data = self.parent.image.get()
                    # pylint: disable=protected-access
                    self.parent._run_subs(sub_type=self.parent.SUB_MONITOR, value=data)
                    if self.parent.stopped:
                        raise DeviceStopError(f"{self.parent.name} was stopped")
                    if self.parent.write_to_disk.get():
                        self.parent.h5_writer.receive_data(data)
                status.set_finished()
            # pylint: disable=broad-except
            except Exception as exc:
                content = traceback.format_exc()
                logger.warning(
                    f"Error in on_trigger_call in device {self.parent.name}. Error traceback: {content}"
                )
                status.set_exception(exc)

        self._thread_trigger = Thread(target=on_trigger_call, args=(status,))
        self._thread_trigger.start()
        return status

    def on_stage(self) -> None:
        """Stage the camera for upcoming scan

        This method is called from BEC in preparation of a scan.
        It receives metadata about the scan from BEC,
        compiles it and prepares the camera for the scan.

        FYI: No data is written to disk in the simulation, but upon each trigger it
        is published to the device_monitor endpoint in REDIS.
        """
        self.parent.filepath.set(
            self.parent.filewriter.compile_full_filename(f"{self.parent.name}")
        ).wait()

        self.parent.frames.set(
            self.parent.scaninfo.num_points * self.parent.scaninfo.frames_per_trigger
        )
        self.parent.exp_time.set(self.parent.scaninfo.exp_time)
        self.parent.burst.set(self.parent.scaninfo.frames_per_trigger)
        if self.parent.write_to_disk.get():
            self.parent.h5_writer.prepare(
                file_path=self.parent.filepath.get(), h5_entry="/entry/data/data"
            )
            self.publish_file_location(done=False, successful=False)
        self.parent.stopped = False

    def on_complete(self) -> None:
        """Complete the motion of the simulated device."""
        status = DeviceStatus(self.parent)

        def on_complete_call(status: DeviceStatus) -> None:
            try:
                if self.parent.write_to_disk.get():
                    self.parent.h5_writer.write_data()
                self.publish_file_location(done=True, successful=True)
                if self.parent.stopped:
                    raise DeviceStopError(f"{self.parent.name} was stopped")
                status.set_finished()
            # pylint: disable=broad-except
            except Exception as exc:
                content = traceback.format_exc()
                logger.warning(
                    f"Error in on_complete call in device {self.parent.name}. Error traceback: {content}"
                )
                status.set_exception(exc)

        self._thread_complete = Thread(target=on_complete_call, args=(status,), daemon=True)
        self._thread_complete.start()
        return status

    def on_stop(self) -> None:
        """Stop the camera acquisition."""
        if self._thread_trigger:
            self._thread_trigger.join()
        if self._thread_complete:
            self._thread_complete.join()
        self._thread_trigger = None
        self._thread_complete = None


class SimCamera(PSIDetectorBase):
    """A simulated device mimic any 2D camera.

    It's image is a computed signal, which is configurable by the user and from the command line.
    The corresponding simulation class is sim_cls=SimulatedDataCamera, more details on defaults within the simulation class.

    >>> camera = SimCamera(name="camera")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argmuent, passed on to all signals of the device.
    precision (integer)     : Precision of the readback in digits, written to .describe(). Default is 3 digits.
    sim_init (dict)         : Dictionary to initiate parameters of the simulation, check simulation type defaults for more details.
    parent                  : Parent device, optional, is used internally if this signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional. Default is Kind.normal. See Kind for options.
    device_manager          : DeviceManager from BEC, optional . Within startup of simulation, device_manager is passed on automatically.

    """

    USER_ACCESS = ["sim", "registered_proxies"]

    custom_prepare_cls = SimCameraSetup
    sim_cls = SimulatedDataCamera
    SHAPE = (100, 100)
    BIT_DEPTH = np.uint16

    SUB_MONITOR = "device_monitor_2d"
    _default_sub = SUB_MONITOR

    exp_time = Cpt(SetableSignal, name="exp_time", value=1, kind=Kind.config)
    file_pattern = Cpt(SetableSignal, name="file_pattern", value="", kind=Kind.config)
    frames = Cpt(SetableSignal, name="frames", value=1, kind=Kind.config)
    burst = Cpt(SetableSignal, name="burst", value=1, kind=Kind.config)

    image_shape = Cpt(SetableSignal, name="image_shape", value=SHAPE, kind=Kind.config)
    image = Cpt(
        ReadOnlySignal,
        name="image",
        value=np.empty(SHAPE, dtype=BIT_DEPTH),
        compute_readback=True,
        kind=Kind.omitted,
    )
    write_to_disk = Cpt(SetableSignal, name="write_to_disk", value=False, kind=Kind.config)

    def __init__(
        self, name, *, kind=None, parent=None, sim_init: dict = None, device_manager=None, **kwargs
    ):
        self.sim_init = sim_init
        self._registered_proxies = {}
        self.sim = self.sim_cls(parent=self, **kwargs)
        self.h5_writer = H5Writer()
        super().__init__(
            name=name, parent=parent, kind=kind, device_manager=device_manager, **kwargs
        )
        if self.sim_init:
            self.sim.set_init(self.sim_init)

    @property
    def registered_proxies(self) -> None:
        """Dictionary of registered signal_names and proxies."""
        return self._registered_proxies



# Description: Launch the scan bundler.
# This script is the entry point for the Scan Bundler. It is called either
# by the bec-scan-bundler entry point or directly from the command line.
import threading

from bec_lib.bec_service import parse_cmdline_args
from bec_lib.logger import bec_logger
from bec_lib.redis_connector import RedisConnector
from bec_server import scan_bundler

logger = bec_logger.logger
bec_logger.level = bec_logger.LOGLEVEL.INFO


def main():
    """
    Launch the scan bundler.
    """
    _, _, config = parse_cmdline_args()

    sb = scan_bundler.ScanBundler(config, RedisConnector)

    try:
        event = threading.Event()
        logger.success("Started ScanBundler")
        event.wait()
    except KeyboardInterrupt:
        sb.shutdown()


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import pytest
from typeguard import TypeCheckError

from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.redis_connector import MessageObject
from bec_lib.scan_manager import ScanManager

if TYPE_CHECKING:
    from bec_lib.redis_connector import RedisConnector


@pytest.fixture
def scan_manager():
    connector = mock.MagicMock()
    manager = ScanManager(connector=connector)
    yield manager
    manager.shutdown()


@pytest.fixture
def scan_manager_with_scan(scan_queue_status_msg):
    connector = mock.MagicMock()
    manager = ScanManager(connector=connector)
    manager.scan_storage.update_with_queue_status(scan_queue_status_msg)
    yield manager
    manager.shutdown()


@pytest.fixture
def scan_manager_with_fakeredis(connected_connector: RedisConnector):
    manager = ScanManager(connector=connected_connector)
    yield manager
    manager.shutdown()


def test_scan_manager_next_scan_number(scan_manager):
    scan_manager.connector.get.return_value = messages.VariableMessage(value=3)
    assert scan_manager.next_scan_number == 3


def test_scan_manager_next_scan_number_failed(scan_manager):
    scan_manager.connector.get.return_value = None
    assert scan_manager.next_scan_number == -1


def test_scan_manager_next_scan_number_with_int(scan_manager):
    scan_manager.connector.get.return_value = 3
    assert scan_manager.next_scan_number == 3


def test_scan_manager_next_scan_number_setter(scan_manager):
    scan_manager.next_scan_number = 3
    scan_manager.connector.set.assert_called_once_with(
        MessageEndpoints.scan_number(), messages.VariableMessage(value=3)
    )


def test_scan_manager_next_dataset_number(scan_manager):
    scan_manager.connector.get.return_value = messages.VariableMessage(value=3)
    assert scan_manager.next_dataset_number == 3


def test_scan_manager_next_dataset_number_failed(scan_manager):
    scan_manager.connector.get.return_value = None
    assert scan_manager.next_dataset_number == -1


def test_scan_manager_next_dataset_number_with_int(scan_manager):
    scan_manager.connector.get.return_value = 3
    assert scan_manager.next_dataset_number == 3


def test_scan_manager_next_dataset_number_setter(scan_manager):
    scan_manager.next_dataset_number = 3
    scan_manager.connector.set.assert_called_once_with(
        MessageEndpoints.dataset_number(), messages.VariableMessage(value=3)
    )


def test_scan_manager_request_scan_abortion(scan_manager):
    scan_manager.request_scan_abortion("scan_id")
    scan_manager.connector.send.assert_called_once_with(
        MessageEndpoints.scan_queue_modification_request(),
        messages.ScanQueueModificationMessage(scan_id="scan_id", action="abort", parameter={}),
    )


@pytest.mark.parametrize("scan_id", [None, "scan_id", ["scan_id"], [None]])
def test_scan_manager_request_scan_abortion_scan_id(scan_manager, scan_id):

    class ScanStorage:
        current_scan_info = {"scan_id": scan_id}

        @property
        def current_scan_id(self):
            return self.current_scan_info["scan_id"]

    scan_manager.scan_storage = ScanStorage()
    scan_manager.request_scan_abortion()
    scan_manager.connector.send.assert_called_once_with(
        MessageEndpoints.scan_queue_modification_request(),
        messages.ScanQueueModificationMessage(scan_id=scan_id, action="abort", parameter={}),
    )


def test_scan_manager_request_scan_halt(scan_manager):
    scan_manager.request_scan_halt("scan_id")
    scan_manager.connector.send.assert_called_once_with(
        MessageEndpoints.scan_queue_modification_request(),
        messages.ScanQueueModificationMessage(scan_id="scan_id", action="halt", parameter={}),
    )


@pytest.mark.parametrize("scan_id", [None, "scan_id", ["scan_id"], [None]])
def test_scan_manager_request_scan_halt_scan_id(scan_manager, scan_id):

    class ScanStorage:
        current_scan_info = {"scan_id": scan_id}

        @property
        def current_scan_id(self):
            return self.current_scan_info["scan_id"]

    scan_manager.scan_storage = ScanStorage()
    scan_manager.request_scan_halt()
    scan_manager.connector.send.assert_called_once_with(
        MessageEndpoints.scan_queue_modification_request(),
        messages.ScanQueueModificationMessage(scan_id=scan_id, action="halt", parameter={}),
    )


def test_scan_manager_request_scan_continuation(scan_manager):
    scan_manager.request_scan_continuation("scan_id")
    scan_manager.connector.send.assert_called_once_with(
        MessageEndpoints.scan_queue_modification_request(),
        messages.ScanQueueModificationMessage(scan_id="scan_id", action="continue", parameter={}),
    )


@pytest.mark.parametrize("scan_id", [None, "scan_id", ["scan_id"], [None]])
def test_scan_manager_request_scan_continuation_scan_id(scan_manager, scan_id):

    class ScanStorage:
        current_scan_info = {"scan_id": scan_id}

        @property
        def current_scan_id(self):
            return self.current_scan_info["scan_id"]

    scan_manager.scan_storage = ScanStorage()
    scan_manager.request_scan_continuation()
    scan_manager.connector.send.assert_called_once_with(
        MessageEndpoints.scan_queue_modification_request(),
        messages.ScanQueueModificationMessage(scan_id=scan_id, action="continue", parameter={}),
    )


def test_scan_manager_add_scan_to_queue_schedule(scan_manager_with_fakeredis):
    """
    Test the interaction with queue schedules

    Args:
        scan_manager_with_fakeredis: The scan manager fixture with a fakeredis connection
    """
    manager: ScanManager = scan_manager_with_fakeredis
    msg = messages.ScanQueueMessage(scan_type="mv", parameter={"args": {"samx": [5], "samy": [5]}})
    manager.add_scan_to_queue_schedule("new_schedule", msg)

    with pytest.raises(TypeCheckError):
        manager.add_scan_to_queue_schedule("new_schedule", {})

    assert manager.get_scan_queue_schedule("new_schedule") == [msg]

    msg2 = messages.ScanQueueMessage(scan_type="mv", parameter={"args": {"samx": [6], "samy": [6]}})
    manager.add_scan_to_queue_schedule("new_schedule", msg2)

    assert manager.get_scan_queue_schedule("new_schedule") == [msg, msg2]

    manager.add_scan_to_queue_schedule("new_schedule2", msg)

    assert manager.get_scan_queue_schedule("new_schedule2") == [msg]

    assert manager.get_scan_queue_schedule_names() == ["new_schedule", "new_schedule2"]

    manager.clear_scan_queue_schedule("new_schedule2")

    assert manager.get_scan_queue_schedule_names() == ["new_schedule"]

    assert manager.get_scan_queue_schedule("new_schedule2") == []

    assert manager.get_scan_queue_schedule("new_schedule") == [msg, msg2]

    manager.clear_all_scan_queue_schedules()

    assert manager.get_scan_queue_schedule_names() == []


def test_scan_manager_add_public_file(scan_manager_with_scan):
    """
    Test the public file callback. It should add the file info to the scan item
    of the file's scan.

    For this, we use the scan_manager_with_scan fixture, which has a scan item
    already in the queue. The queue fixture is defined in conftest.py.
    """
    msg = messages.FileMessage(
        file_path="/Users/scans/S00001_master.h5", done=True, successful=True
    )
    msg_object = MessageObject(
        topic=MessageEndpoints.public_file(
            "bfa582aa-f9cd-4258-ab5d-3e5d54d3dde5", "master"
        ).endpoint,
        value=msg,
    )
    # pylint: disable=protected-access
    scan_manager_with_scan._public_file_callback(msg=msg_object)
    assert scan_manager_with_scan.scan_storage.storage[-1].public_files == {
        msg.file_path: {"done_state": True, "successful": True}
    }
    assert (
        "File: /Users/scans/S00001_master.h5"
        in scan_manager_with_scan.scan_storage.storage[-1].describe()
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from bec_lib.devicemanager import DeviceManagerBase
    from bec_server.file_writer.file_writer import HDF5Storage


class DefaultFormat:
    """
    Default NeXus file format.
    """

    def __init__(
        self,
        storage: HDF5Storage,
        data: dict,
        info_storage: dict,
        file_references: dict,
        device_manager: DeviceManagerBase,
    ):
        self.storage = storage
        self.data = data
        self.file_references = file_references
        self.device_manager = device_manager
        self.info_storage = info_storage

    def get_storage_format(self) -> dict:
        """
        Internal method to extract the storage format after formatting the data. This method
        should not be called directly.

        Returns:
            dict: The storage format.
        """
        self.format()
        # pylint: disable=protected-access
        return self.storage._storage

    def get_entry(self, name: str, default=None) -> Any:
        """
        Get an entry from the scan data assuming a <device>.<device>.value structure.

        This method is a helper to extract the device data from the scan data, irrespective of the
        data structure (list of entries or single entry).

        Args:
            name (str): Entry name
            default (Any, optional): Default value. Defaults to None.
        """
        if isinstance(self.data.get(name), list) and isinstance(self.data.get(name)[0], dict):
            return [
                sub_data.get(name, {}).get("value", default) for sub_data in self.data.get(name)
            ]

        return self.data.get(name, {}).get(name, {}).get("value", default)

    def format(self) -> None:
        """
        Prepare the NeXus file format.
        Override this method in file writer plugins to customize the HDF5 file format.

        The class provides access to the following attributes:
        - self.storage: The HDF5Storage object.
        - self.data: The data dictionary.
        - self.file_references: The file references dictionary.
        - self.device_manager: The DeviceManagerBase object.

        See also: :class:`bec_server.file_writer.file_writer.HDF5Storage`.

        """
        # /entry
        entry = self.storage.create_group("entry")
        entry.attrs["NX_class"] = "NXentry"
        entry.attrs["start_time"] = self.info_storage.get("start_time")
        entry.attrs["end_time"] = self.info_storage.get("end_time")
        entry.attrs["version"] = 1.0

        # /entry/collection
        collection = entry.create_group("collection")
        collection.attrs["NX_class"] = "NXcollection"
        devices = collection.create_dataset("devices", data=self.data)
        devices.attrs["NX_class"] = "NXcollection"
        metadata = collection.create_dataset("metadata", data=self.info_storage)
        metadata.attrs["NX_class"] = "NXcollection"

        # /entry/control
        control = entry.create_group("control")
        control.attrs["NX_class"] = "NXmonitor"
        control.create_dataset(name="mode", data="monitor")

        # /entry/data
        if "eiger_4" in self.device_manager.devices:
            entry.create_soft_link(name="data", target="/entry/instrument/eiger_4")

        # /entry/sample
        control = entry.create_group("sample")
        control.attrs["NX_class"] = "NXsample"
        control.create_dataset(name="name", data=self.data.get("samplename"))
        control.create_dataset(name="description", data=self.data.get("sample_description"))

        # /entry/instrument
        instrument = entry.create_group("instrument")
        instrument.attrs["NX_class"] = "NXinstrument"

        source = instrument.create_group("source")
        source.attrs["NX_class"] = "NXsource"
        source.create_dataset(name="type", data="Synchrotron X-ray Source")
        source.create_dataset(name="name", data="Swiss Light Source")
        source.create_dataset(name="probe", data="x-ray")



from unittest import mock

import pytest

from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_server.scan_bundler.bec_emitter import BECEmitter


@pytest.fixture
def bec_emitter_mock(scan_bundler_mock):
    with mock.patch.object(BECEmitter, "_start_buffered_connector") as start:
        emitter = BECEmitter(scan_bundler_mock)
        start.assert_called_once()
        with mock.patch.object(emitter, "_get_messages_from_buffer"):
            yield emitter
    emitter.shutdown()


def test_on_scan_point_emit_BEC(bec_emitter_mock):
    sb = bec_emitter_mock.scan_bundler
    with mock.patch.object(bec_emitter_mock, "_send_bec_scan_point") as send:
        bec_emitter_mock.on_scan_point_emit("scan_id", 2)
        send.assert_called_once_with("scan_id", 2)


def test_on_baseline_emit_BEC(bec_emitter_mock):
    sb = bec_emitter_mock.scan_bundler
    with mock.patch.object(bec_emitter_mock, "_send_baseline") as send:
        bec_emitter_mock.on_baseline_emit("scan_id")
        send.assert_called_once_with("scan_id")


def test_send_bec_scan_point(bec_emitter_mock):
    sb = bec_emitter_mock.scan_bundler
    scan_id = "lkajsdlkj"
    point_id = 2
    sb.sync_storage[scan_id] = {"info": {}, "status": "open", "sent": set()}
    sb.sync_storage[scan_id][point_id] = {}
    msg = messages.ScanMessage(
        point_id=point_id,
        scan_id=scan_id,
        data=sb.sync_storage[scan_id][point_id],
        metadata={"scan_id": "lkajsdlkj", "scan_type": None, "scan_report_devices": None},
    )
    with mock.patch.object(bec_emitter_mock, "add_message") as send:
        bec_emitter_mock._send_bec_scan_point(scan_id, point_id)
        send.assert_called_once_with(
            msg,
            MessageEndpoints.scan_segment(),
            MessageEndpoints.public_scan_segment(scan_id, point_id),
        )


def test_send_baseline_BEC(bec_emitter_mock):
    sb = bec_emitter_mock.scan_bundler
    scan_id = "lkajsdlkj"
    sb.sync_storage[scan_id] = {"info": {}, "status": "open", "sent": set()}
    sb.sync_storage[scan_id]["baseline"] = {}
    msg = messages.ScanBaselineMessage(scan_id=scan_id, data=sb.sync_storage[scan_id]["baseline"])
    with mock.patch.object(sb, "connector") as connector:
        bec_emitter_mock._send_baseline(scan_id)
        pipe = connector.pipeline()
        connector.set.assert_called_once_with(
            MessageEndpoints.public_scan_baseline(scan_id), msg, expire=1800, pipe=pipe
        )
        connector.set_and_publish.assert_called_once_with(
            MessageEndpoints.scan_baseline(), msg, pipe=pipe
        )


@pytest.mark.parametrize(
    "msgs",
    [
        ([]),
        (
            [
                (
                    messages.ScanMessage(point_id=1, scan_id="scan_id", data={}, metadata={}),
                    "endpoint",
                    None,
                )
            ]
        ),
        (
            [
                (
                    messages.ScanMessage(point_id=1, scan_id="scan_id", data={}, metadata={}),
                    "endpoint",
                    None,
                ),
                (
                    messages.ScanMessage(point_id=2, scan_id="scan_id", data={}, metadata={}),
                    "endpoint",
                    None,
                ),
            ]
        ),
        (
            [
                (
                    messages.ScanMessage(point_id=1, scan_id="scan_id", data={}, metadata={}),
                    "endpoint",
                    "public_endpoint",
                ),
                (
                    messages.ScanMessage(point_id=2, scan_id="scan_id", data={}, metadata={}),
                    "endpoint",
                    "public_endpoint",
                ),
            ]
        ),
    ],
)
def test_publish_data(msgs, bec_emitter_mock):
    connector = bec_emitter_mock.connector = mock.MagicMock()
    bec_emitter_mock._get_messages_from_buffer.return_value = msgs
    bec_emitter_mock._publish_data()
    bec_emitter_mock._get_messages_from_buffer.assert_called_once()

    if not msgs:
        connector.send.assert_not_called()
        return

    pipe = connector.pipeline()
    msgs_bundle = messages.BundleMessage()
    _, endpoint, _ = msgs[0]
    for msg, endpoint, public in msgs:
        msg_dump = msg
        msgs_bundle.append(msg_dump)
        if public:
            connector.set.assert_has_calls(connector.set(public, msg_dump, pipe=pipe, expire=1800))

    connector.send.assert_called_with(endpoint, msgs_bundle, pipe=pipe)


@pytest.mark.parametrize(
    "msg,endpoint,public",
    [
        (
            messages.ScanMessage(point_id=1, scan_id="scan_id", data={}, metadata={}),
            "endpoint",
            None,
        ),
        (
            messages.ScanMessage(point_id=1, scan_id="scan_id", data={}, metadata={}),
            "endpoint",
            "public",
        ),
    ],
)
def test_add_message(msg, endpoint, public):
    connector = mock.MagicMock()
    emitter = BECEmitter(connector)
    emitter.add_message(msg, endpoint, public)
    msgs = emitter._get_messages_from_buffer()
    out_msg, out_endpoint, out_public = msgs[0]
    assert out_msg == msg
    assert out_endpoint == endpoint
    assert out_public == public
    emitter.shutdown()



from unittest import mock

import pytest

from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_server.scan_server.device_validation import DeviceValidation
from bec_server.scan_server.errors import ScanAbortion


@pytest.fixture
def validator():
    connector = mock.MagicMock()
    worker = mock.MagicMock()
    yield DeviceValidation(connector, worker)


def test_device_validator_devices_are_ready(validator):
    with mock.patch.object(validator, "get_device_status") as mock_get_device_status:
        mock_get_device_status.return_value = [
            messages.DeviceReqStatusMessage(
                device="dev1",
                success=True,
                metadata={"action": "complete", "scan_id": "scan_id", "RID": "RID", "DIID": 20},
            )
        ]
        assert validator.devices_are_ready(
            devices=["dev1"],
            endpoint=MessageEndpoints.device_req_status,
            message_cls=messages.DeviceReqStatusMessage,
            metadata={"scan_id": "scan_id", "RID": "RID", "DIID": 20},
            checks=[validator.devices_returned_successfully],
        )


def test_device_validator_wait_until_ready(validator):
    round_1 = [
        messages.DeviceReqStatusMessage(
            device="dev1",
            success=True,
            metadata={"action": "complete", "scan_id": "old_scan_id", "RID": "old_RID", "DIID": 20},
        )
    ]
    round_2 = [
        messages.DeviceReqStatusMessage(
            device="dev1",
            success=True,
            metadata={"action": "complete", "scan_id": "scan_id", "RID": "RID", "DIID": 20},
        )
    ]
    with mock.patch.object(validator, "get_device_status") as mock_get_device_status:
        mock_get_device_status.side_effect = [round_1, round_2]

        assert (
            validator.devices_are_ready(
                devices=["dev1"],
                endpoint=MessageEndpoints.device_req_status,
                message_cls=messages.DeviceReqStatusMessage,
                metadata={"scan_id": "scan_id", "RID": "RID", "DIID": 20},
                checks=[validator.devices_returned_successfully],
            )
            is False
        )

        assert (
            validator.devices_are_ready(
                devices=["dev1"],
                endpoint=MessageEndpoints.device_req_status,
                message_cls=messages.DeviceReqStatusMessage,
                metadata={"scan_id": "scan_id", "RID": "RID", "DIID": 20},
                checks=[validator.devices_returned_successfully],
            )
            is True
        )


def test_device_validator_raises_on_failed_status(validator):
    round_1 = [
        messages.DeviceReqStatusMessage(
            device="dev1",
            success=True,
            metadata={"action": "complete", "scan_id": "old_scan_id", "RID": "old_RID", "DIID": 20},
        )
    ]
    round_2 = [
        messages.DeviceReqStatusMessage(
            device="dev1",
            success=False,
            metadata={"action": "complete", "scan_id": "scan_id", "RID": "RID", "DIID": 20},
        )
    ]
    with mock.patch.object(validator, "get_device_status") as mock_get_device_status:
        mock_get_device_status.side_effect = [round_1, round_2]

        assert (
            validator.devices_are_ready(
                devices=["dev1"],
                endpoint=MessageEndpoints.device_req_status,
                message_cls=messages.DeviceReqStatusMessage,
                metadata={"scan_id": "scan_id", "RID": "RID", "DIID": 20},
                checks=[validator.devices_returned_successfully],
            )
            is False
        )
        with pytest.raises(ScanAbortion):
            validator.devices_are_ready(
                devices=["dev1"],
                endpoint=MessageEndpoints.device_req_status,
                message_cls=messages.DeviceReqStatusMessage,
                metadata={"scan_id": "scan_id", "RID": "RID", "DIID": 20},
                checks=[validator.devices_returned_successfully],
            )



# This file is the entry point for the BEC device server. It is responsible for
# launching the device server and handling command line arguments.
# It is called either by the bec-device-server entry point or directly from the command line.

# pylint: disable=wrong-import-position
# pylint: disable=wrong-import-order
# we need to run the startup script before we import anything else. This is
# to ensure that the epics environment variables are set correctly.
import importlib.metadata as imd

entry_points = imd.entry_points(group="bec.deployment.device_server")
for entry_point in entry_points:
    if entry_point.name == "plugin_ds_startup":
        entry_point.load()()

import argparse
import threading

from bec_lib.bec_service import parse_cmdline_args
from bec_lib.logger import bec_logger
from bec_lib.redis_connector import RedisConnector
from bec_server import device_server

logger = bec_logger.logger
bec_logger.level = bec_logger.LOGLEVEL.INFO


def main():
    """
    Launch the BEC device server.
    """
    _, _, config = parse_cmdline_args()

    s = device_server.DeviceServer(config, RedisConnector)
    try:
        event = threading.Event()
        s.start()
        logger.success("Started DeviceServer")
        event.wait()
    except KeyboardInterrupt:
        s.shutdown()


if __name__ == "__main__":
    main()



""" This module tests the bec_lib.client module. """

import pytest

from bec_lib.client import SystemConfig


def test_system_config():
    """Test the SystemConfig class."""
    config = SystemConfig(file_suffix="suff", file_directory="dir")
    assert config.file_suffix == "suff"
    assert config.file_directory == "dir"
    config = SystemConfig()
    assert config.file_suffix is None
    assert config.file_directory is None
    config.file_suffix = "suff_-"
    config.file_directory = "/dir_-/blabla"
    assert config.file_suffix == "suff_-"
    assert config.file_directory == "dir_-/blabla"
    with pytest.raises(ValueError):
        config = SystemConfig(file_suffix="@")
        config = SystemConfig(file_directory="ä")



# Description: Launch the data processing server.
# This script is the entry point for the Data Processing Server. It is called either
# by the bec-dap entry point or directly from the command line.
import argparse
import threading

import bec_server.data_processing as data_processing
from bec_lib.bec_service import parse_cmdline_args
from bec_lib.logger import bec_logger
from bec_lib.redis_connector import RedisConnector
from bec_server.data_processing.lmfit1d_service import LmfitService1D

logger = bec_logger.logger
bec_logger.level = bec_logger.LOGLEVEL.INFO


def main():
    """
    Launch the data processing server.
    """
    _, _, config = parse_cmdline_args()

    bec_server = data_processing.dap_server.DAPServer(
        config=config, connector_cls=RedisConnector, provided_services=[LmfitService1D]
    )
    bec_server.start()

    try:
        event = threading.Event()
        logger.success(
            f"Started DAP server for {bec_server._service_id} services. Press Ctrl+C to stop."
        )
        event.wait()
    except KeyboardInterrupt:
        bec_server.shutdown()
        event.set()


if __name__ == "__main__":
    main()



"""Module for a simulated 1D Waveform detector, i.e. a Falcon XRF detector."""

import os
import threading
import traceback

import numpy as np
from bec_lib.logger import bec_logger
from ophyd import Component as Cpt
from ophyd import Device, DeviceStatus, Kind

from ophyd_devices.sim.sim_data import SimulatedDataWaveform
from ophyd_devices.sim.sim_signals import ReadOnlySignal, SetableSignal
from ophyd_devices.utils.bec_scaninfo_mixin import BecScaninfoMixin
from ophyd_devices.utils.errors import DeviceStopError

logger = bec_logger.logger


class SimWaveform(Device):
    """A simulated device mimic any 1D Waveform detector.

    It's waveform is a computed signal, which is configurable by the user and from the command line.
    The corresponding simulation class is sim_cls=SimulatedDataWaveform, more details on defaults within the simulation class.

    >>> waveform = SimWaveform(name="waveform")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argmuent, passed on to all signals of the device.
    precision (integer)     : Precision of the readback in digits, written to .describe(). Default is 3 digits.
    sim_init (dict)         : Dictionary to initiate parameters of the simulation, check simulation type defaults for more details.
    parent                  : Parent device, optional, is used internally if this signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional. Default is Kind.normal. See Kind for options.
    device_manager          : DeviceManager from BEC, optional . Within startup of simulation, device_manager is passed on automatically.

    """

    USER_ACCESS = ["sim", "registered_proxies"]

    sim_cls = SimulatedDataWaveform
    SHAPE = (1000,)
    BIT_DEPTH = np.uint16

    SUB_MONITOR = "monitor"
    _default_sub = SUB_MONITOR

    exp_time = Cpt(SetableSignal, name="exp_time", value=1, kind=Kind.config)
    file_path = Cpt(SetableSignal, name="file_path", value="", kind=Kind.config)
    file_pattern = Cpt(SetableSignal, name="file_pattern", value="", kind=Kind.config)
    frames = Cpt(SetableSignal, name="frames", value=1, kind=Kind.config)
    burst = Cpt(SetableSignal, name="burst", value=1, kind=Kind.config)

    waveform_shape = Cpt(SetableSignal, name="waveform_shape", value=SHAPE, kind=Kind.config)
    waveform = Cpt(
        ReadOnlySignal,
        name="waveform",
        value=np.empty(SHAPE, dtype=BIT_DEPTH),
        compute_readback=True,
        kind=Kind.omitted,
    )

    def __init__(
        self, name, *, kind=None, parent=None, sim_init: dict = None, device_manager=None, **kwargs
    ):
        self.device_manager = device_manager
        self.sim_init = sim_init
        self._registered_proxies = {}
        self.sim = self.sim_cls(parent=self, **kwargs)

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        self.stopped = False
        self._staged = False
        self.scaninfo = None
        self._trigger_thread = None
        self._update_scaninfo()
        if self.sim_init:
            self.sim.set_init(self.sim_init)

    @property
    def registered_proxies(self) -> None:
        """Dictionary of registered signal_names and proxies."""
        return self._registered_proxies

    def trigger(self) -> DeviceStatus:
        """Trigger the camera to acquire images.

        This method can be called from BEC during a scan. It will acquire images and send them to BEC.
        Whether the trigger is send from BEC is determined by the softwareTrigger argument in the device config.

        Here, we also run a callback on SUB_MONITOR to send the image data the device_monitor endpoint in BEC.
        """
        status = DeviceStatus(self)

        def acquire(status: DeviceStatus):
            try:
                for _ in range(self.burst.get()):
                    self._run_subs(sub_type=self.SUB_MONITOR, value=self.waveform.get())
                    if self.stopped:
                        raise DeviceStopError(f"{self.name} was stopped")
                status.set_finished()
            # pylint: disable=broad-except
            except Exception as exc:
                content = traceback.format_exc()
                status.set_exception(exc=exc)
                logger.warning(f"Error in {self.name} trigger; Traceback: {content}")

        self._trigger_thread = threading.Thread(target=acquire, args=(status,), daemon=True)
        self._trigger_thread.start()
        return status

    def _update_scaninfo(self) -> None:
        """Update scaninfo from BecScaninfoMixing
        This depends on device manager and operation/sim_mode
        """
        self.scaninfo = BecScaninfoMixin(self.device_manager)

    def stage(self) -> list[object]:
        """Stage the camera for upcoming scan

        This method is called from BEC in preparation of a scan.
        It receives metadata about the scan from BEC,
        compiles it and prepares the camera for the scan.

        FYI: No data is written to disk in the simulation, but upon each trigger it
        is published to the device_monitor endpoint in REDIS.
        """
        if self._staged:
            return super().stage()
        self.scaninfo.load_scan_metadata()
        self.file_path.set(
            os.path.join(
                self.file_path.get(), self.file_pattern.get().format(self.scaninfo.scan_number)
            )
        )
        self.frames.set(self.scaninfo.num_points * self.scaninfo.frames_per_trigger)
        self.exp_time.set(self.scaninfo.exp_time)
        self.burst.set(self.scaninfo.frames_per_trigger)
        self.stopped = False
        return super().stage()

    def unstage(self) -> list[object]:
        """Unstage the device

        Send reads from all config signals to redis
        """
        if self.stopped is True or not self._staged:
            return super().unstage()

        return super().unstage()

    def stop(self, *, success=False):
        """Stop the device"""
        self.stopped = True
        if self._trigger_thread:
            self._trigger_thread.join()
        self._trigger_thread = None
        super().stop(success=success)



from __future__ import annotations

import datetime
import json
import os
import traceback
import typing

import h5py

from bec_lib import messages, plugin_helper
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger

from .default_writer import DefaultFormat as default_NeXus_format
from .merged_dicts import merge_dicts

logger = bec_logger.logger


class NeXusLayoutError(Exception):
    """
    Exception raised when the NeXus layout is incorrect.
    """


class HDF5Storage:
    """
    The HDF5Storage class is a container used by the HDF5 writer plugins to store data in the correct NeXus format.
    """

    def __init__(self, storage_type: str = "group", data=None) -> None:
        self._storage = {}
        self._storage_type = storage_type
        self.attrs = {}
        self._data = data

    def create_group(self, name: str) -> HDF5Storage:
        """
        Create a group in the HDF5 storage.

        Args:
            name (str): Group name

        Returns:
            HDF5Storage: Group storage
        """
        self._storage[name] = HDF5Storage(storage_type="group")
        return self._storage[name]

    def create_dataset(self, name: str, data: typing.Any) -> HDF5Storage:
        """
        Create a dataset in the HDF5 storage.

        Args:
            name (str): Dataset name
            data (typing.Any): Dataset data

        Returns:
            HDF5Storage: Dataset storage
        """
        self._storage[name] = HDF5Storage(storage_type="dataset", data=data)
        return self._storage[name]

    def create_soft_link(self, name: str, target: str) -> HDF5Storage:
        """
        Create a soft link in the HDF5 storage.

        Args:
            name (str): Link name
            target (str): Link target

        Returns:
            HDF5Storage: Link storage
        """
        self._storage[name] = HDF5Storage(storage_type="softlink", data=target)
        return self._storage[name]

    def create_ext_link(self, name: str, target: str, entry: str) -> HDF5Storage:
        """
        Create an external link in the HDF5 storage.

        Args:
            name (str): Link name
            target (str): Name of the target file
            entry (str): Entry within the target file (e.g. entry/instrument/eiger_4)

        Returns:
            HDF5Storage: Link storage
        """
        data = {"file": target, "entry": entry}
        self._storage[name] = HDF5Storage(storage_type="ext_link", data=data)
        return self._storage[name]


class HDF5StorageWriter:
    """
    The HDF5StorageWriter class is used to write the HDF5Storage object to an HDF5 file.

    The class
    """

    device_storage = None
    info_storage = None

    def add_group(self, name: str, container: typing.Any, val: HDF5Storage):
        group = container.create_group(name)
        self.add_attribute(group, val.attrs)
        self.add_content(group, val._storage)

        data = val._data

        if not data:
            return

        for key, value in data.items():
            if value is None:
                continue
            if isinstance(value, dict):
                sub_storage = HDF5Storage(key)
                dict_to_storage(sub_storage, value)
                self.add_group(key, group, sub_storage)
                # self.add_content(group, sub_storage._storage)
                continue
            if isinstance(value, list) and isinstance(value[0], dict):
                merged_dict = merge_dicts(value)
                sub_storage = HDF5Storage(key)
                dict_to_storage(sub_storage, merged_dict)
                self.add_group(key, group, sub_storage)
                continue

            group.create_dataset(name=key, data=value)

    def add_dataset(self, name: str, container: typing.Any, val: HDF5Storage):
        try:
            if isinstance(val._data, dict):
                self.add_group(name, container, val)
                return

            data = val._data
            if data is None:
                return
            if isinstance(data, list):
                if data and isinstance(data[0], dict):
                    data = json.dumps(data)
            dataset = container.create_dataset(name, data=data)
            self.add_attribute(dataset, val.attrs)
            self.add_content(dataset, val._storage)
        except Exception:
            content = traceback.format_exc()
            logger.error(f"Failed to write dataset {name}: {content}")
        return

    def add_attribute(self, container: typing.Any, attributes: dict):
        for name, value in attributes.items():
            if value is not None:
                container.attrs[name] = value

    def add_hardlink(self, name, container, val):
        pass

    def add_softlink(self, name, container, val):
        container[name] = h5py.SoftLink(val._data)

    def add_external_link(self, name, container, val):
        container[name] = h5py.ExternalLink(val._data.get("file"), val._data.get("entry"))

    def add_content(self, container, storage):
        for name, val in storage.items():
            if val._storage_type == "group":
                self.add_group(name, container, val)
            elif val._storage_type == "dataset":
                self.add_dataset(name, container, val)
            elif val._storage_type == "hardlink":
                self.add_hardlink(name, container, val)
            elif val._storage_type == "softlink":
                self.add_softlink(name, container, val)
            elif val._storage_type == "ext_link":
                self.add_external_link(name, container, val)
            else:
                pass

    @classmethod
    def write(cls, writer_storage, file):
        writer = cls()
        writer.add_content(file, writer_storage)


class HDF5FileWriter:
    """
    The HDF5FileWriter class is used to write data to an HDF5 file. Internally, it uses the HDF5StorageWriter class to
    write the HDF5Storage object to the file.

    Its primary purpose is to prepare the data, select the correct writer plugin and initiate the writing process.
    """

    def __init__(self, file_writer_manager):
        self.file_writer_manager = file_writer_manager

    @staticmethod
    def _create_device_data_storage(data):
        device_storage = {}
        if data.baseline:
            device_storage.update(data.baseline)
        if data.async_data:
            device_storage.update(data.async_data)
        keys = list(data.scan_segments.keys())
        keys.sort()
        for point in keys:
            for dev in data.scan_segments[point]:
                if dev not in device_storage:
                    device_storage[dev] = [data.scan_segments[point][dev]]
                    continue
                device_storage[dev].append(data.scan_segments[point][dev])
        return device_storage

    def write(self, file_path: str, data):
        """
        Write the data to an HDF5 file.

        Args:
            file_path (str): File path
            data (ScanStorage): Scan data

        Raises:
            NeXusLayoutError: Raised when the NeXus layout is incorrect.
        """
        device_storage = self._create_device_data_storage(data)
        info_storage = {}
        info_storage["bec"] = data.metadata

        # NeXus needs start_time and end_time in ISO8601 format, so we have to convert it
        if data.start_time is not None:
            info_storage["start_time"] = datetime.datetime.fromtimestamp(
                data.start_time
            ).isoformat()
        if data.end_time is not None:
            info_storage["end_time"] = datetime.datetime.fromtimestamp(data.end_time).isoformat()
        info_storage.update(info_storage["bec"].get("user_metadata", {}))
        info_storage["bec"].pop("user_metadata", None)

        requested_plugin = self.file_writer_manager.file_writer_config.get("plugin")
        if requested_plugin == "default_NeXus_format":
            writer_format_cls = default_NeXus_format
        else:
            plugins = plugin_helper.get_file_writer_plugins()
            if requested_plugin not in plugins:
                logger.error(f"Plugin {requested_plugin} not found. Using default plugin.")
                writer_format_cls = default_NeXus_format
            else:
                writer_format_cls = plugins[requested_plugin]

        for file_ref in data.file_references.values():
            rel_path = os.path.relpath(file_ref["path"], os.path.dirname(file_path))
            file_ref["path"] = rel_path

        writer_storage = writer_format_cls(
            storage=HDF5Storage(),
            data=device_storage,
            info_storage=info_storage,
            file_references=data.file_references,
            device_manager=self.file_writer_manager.device_manager,
        ).get_storage_format()

        file_data = {}
        for key, val in device_storage.items():
            file_data[key] = val if not isinstance(val, list) else merge_dicts(val)
        msg_data = {"file_path": file_path, "data": file_data, "scan_info": info_storage}
        msg = messages.FileContentMessage(**msg_data)
        self.file_writer_manager.connector.set_and_publish(MessageEndpoints.file_content(), msg)

        with h5py.File(file_path, "w") as file:
            HDF5StorageWriter.write(writer_storage, file)


def dict_to_storage(storage, data):
    for key, val in data.items():
        if isinstance(val, dict):
            sub = storage.create_group(key)
            dict_to_storage(sub, val)
            continue
        storage.create_dataset(key, val)



from __future__ import annotations

import subprocess
import sys
from typing import TYPE_CHECKING

from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger

if TYPE_CHECKING:
    from bec_lib.redis_connector import RedisConnector

logger = bec_logger.logger


class ServiceHandler:
    def __init__(self, connector: RedisConnector) -> None:
        self.connector = connector
        self.command = f"{sys.executable} -m bec_server.bec_server_utils.launch"

    def start(self):
        self.connector.register(
            MessageEndpoints.service_request(), cb=self.handle_service_request, parent=self
        )

    @staticmethod
    def handle_service_request(
        message: messages.ServiceRequestMessage, parent: ServiceHandler
    ) -> None:
        message = message.value
        if message.action == "restart":
            parent.on_restart()

    def on_restart(self):
        logger.info("Restarting services through service handler")
        subprocess.run(f"{self.command} restart", shell=True, check=True, start_new_session=True)



from bec_lib.device import Device
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger

logger = bec_logger.logger


class DeviceMonitorPlugin:

    def __init__(self, connector):
        self._connector = connector

    def _check_device_avail_for_endpoint(
        self, device_name: str, endpoint: MessageEndpoints
    ) -> bool:
        """Check if the device is available for the given endpoint.

        Args:
            device_name (str): Name of the device
            endpoint (str): Endpoint to check

        Returns:
            bool: True if device is available for the endpoint
        """
        entries = self._connector.keys(f"{endpoint('').endpoint}*")
        avail_devices = [entry.decode().split("/")[-1] for entry in entries]
        return device_name in avail_devices

    def get_data(self, device: str | Device, count: int) -> list:
        """Load the last <count> entries of the device data monitor stream.

        Args:
            device_name (str | Device): Device or name of the device
            count (int): number of images to retrieve

        Returns:
            list: List of numpy arrays
        """
        if isinstance(device, Device):
            device = device.name
        if not self._check_device_avail_for_endpoint(device, MessageEndpoints.device_monitor_2d):
            logger.warning(
                f"Device {device} not available for endpoint MessageEndpoints.device_monitor_2d. Returning None."
            )
            return None
        msgs = self._connector.get_last(MessageEndpoints.device_monitor_2d(device), count=count)
        if msgs is None:
            logger.warning(f"No data found for device {device}. Returning None.")
            return None
        if not isinstance(msgs, list):
            msgs = [msgs]
        im = [sub_msg["data"].data for sub_msg in msgs]
        return im

    def get_data_for_scan(self, device: str | Device, scan: str | int) -> list:
        """Load all available images from the monitor endpoint for a given scan.

        Args:
            device (str | Device): Device or name of the device
            scan (str | int): scan id as string or scan_number as int

        Returns:
            list: List of numpy arrays
        """
        scan_id = None
        if isinstance(device, Device):
            device = device.name
        if not self._check_device_avail_for_endpoint(device, MessageEndpoints.device_monitor_2d):
            logger.warning(
                f"Device {device} not available for endpoint MessageEndpoints.device_monitor_2d. Returning None."
            )
            return None
        msgs = self._connector.xrange(MessageEndpoints.device_monitor_2d(device), min="-", max="+")
        if msgs is None:
            logger.warning(f"No data found for device {device}. Returning None.")
            return None

        if isinstance(scan, int):
            queue_msgs = self._connector.lrange(MessageEndpoints.scan_queue_history(), 0, -1)
            for msg in queue_msgs:
                if msg.content["info"]["scan_number"][0] == scan:
                    scan_id = msg.content["info"]["scan_id"][0]
                    break
            if scan_id is None:
                logger.warning(
                    f"No scan found with scan_number {scan} in queue history. Returning None."
                )
                return None
        elif isinstance(scan, str):
            scan_id = scan
        else:
            raise ValueError(
                f"Value for scan: {scan} must be either scan_number (int) or scan_id (str)"
            )
        im = [
            sub_img["data"].data
            for sub_img in msgs
            if sub_img["data"].metadata.get("scan_id") == scan_id
        ]
        if len(im) == 0:
            logger.warning(
                f"No data found for scan_id: {scan_id} on device {device}. Returning None."
            )
            return None
        return im



"""
This module provides the BECService class, which is the base class for all BEC services.
"""

from __future__ import annotations

import argparse
import getpass
import os
import socket
import threading
import time
import uuid
from dataclasses import asdict, dataclass
from importlib.metadata import version as importlib_version
from typing import TYPE_CHECKING, Any

import psutil
import tomli
from rich.console import Console
from rich.table import Table

from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.service_config import ServiceConfig
from bec_lib.utils.import_utils import lazy_import, lazy_import_from

if TYPE_CHECKING:
    from bec_lib.connector import ConnectorBase

# TODO: put back normal imports when Pydantic gets faster
messages = lazy_import("bec_lib.messages")
BECStatus = lazy_import_from("bec_lib.messages", ("BECStatus",))

logger = bec_logger.logger

SERVICE_CONFIG = None


def parse_cmdline_args(parser=None):
    """Return (parsed args, extra args, ServiceConfig object), the latter being initialized from command line arguments

    Extra args are those passed on the command line, which are not known by the parser

    Arguments:
      - parser (optional): an existing argparse arguments parser, which will be extended with default BECService options
    """
    if parser is None:
        parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--config", default="", help="path to the config file")
    parser.add_argument(
        "--log-level",
        default=None,
        choices=["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"],
        help=f"log level (default: {bec_logger.level.name})",
    )
    parser.add_argument(
        "--file-log-level",
        default=None,
        choices=["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"],
        help="log file log level (not set means 'same as --log-level')",
    )
    parser.add_argument(
        "--redis-log-level",
        default=None,
        choices=["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"],
        help="redis log level (not set means 'same as --log-level')",
    )

    args, extra_args = parser.parse_known_args()

    bec_logger._stderr_log_level = (
        bec_logger.level.name if args.log_level is None else args.log_level
    )
    bec_logger._file_log_level = (
        bec_logger._stderr_log_level if args.file_log_level is None else args.file_log_level
    )
    bec_logger._redis_log_level = (
        bec_logger._stderr_log_level if args.redis_log_level is None else args.redis_log_level
    )

    config_file = args.config
    if config_file:
        service_config = ServiceConfig(config_file, cmdline_args=vars(args))
    else:
        service_config = ServiceConfig(cmdline_args=vars(args))

    return args, extra_args, service_config


class BECService:
    def __init__(
        self,
        config: str | ServiceConfig,
        connector_cls: ConnectorBase,
        unique_service=False,
        wait_for_server=False,
        name: str | None = None,
    ) -> None:
        super().__init__()
        self._name = name if name else self.__class__.__name__
        self._import_config(config)
        self._connector_cls = connector_cls
        self.connector = connector_cls(self.bootstrap_server)
        self._unique_service = unique_service
        self.wait_for_server = wait_for_server
        self.__service_id = str(uuid.uuid4())
        self._user = getpass.getuser()
        self._hostname = socket.gethostname()
        self._service_info_thread = None
        self._service_info_event = threading.Event()
        self._metrics_emitter_thread = None
        self._metrics_emitter_event = threading.Event()
        self._services_info = {}
        self._services_metric = {}
        self._initialize_logger()
        self._check_services()
        self._status = BECStatus.BUSY
        self._start_update_service_info()
        self._start_metrics_emitter()
        self._wait_for_server()
        self._version = None

    @property
    def _service_name(self):
        return self._name if self._unique_service else f"{self._name}/{self._service_id}"

    @property
    def _service_id(self):
        return self.__service_id

    def _import_config(self, config: str | ServiceConfig) -> None:
        if isinstance(config, str):
            self._service_config = ServiceConfig(config_path=config)
        elif isinstance(config, ServiceConfig):
            self._service_config = config
        else:
            raise TypeError("config must be of type str or ServiceConfig")
        global SERVICE_CONFIG
        SERVICE_CONFIG = self._service_config
        self.bootstrap_server = self._service_config.redis

    def _check_services(self, timeout_time=8, sleep_time=0.5) -> None:
        if not self._unique_service:
            return
        elapsed_time = 0
        while self._run_service_check(timeout_time, elapsed_time):
            elapsed_time += sleep_time
            time.sleep(sleep_time)

    def _run_service_check(self, timeout_time: float, elapsed_time: float) -> bool:
        self._update_existing_services()
        try:
            for service_name, msg in self._services_info.items():
                if service_name == self.__class__.__name__:
                    raise RuntimeError(
                        f"Another instance of {self.__class__.__name__} launched by user"
                        f" {msg.content['info']['user']} is already running on"
                        f" {msg.content['info']['hostname']}"
                    )
            return False
        except RuntimeError as service_error:
            if elapsed_time > timeout_time:
                raise RuntimeError from service_error
        return True

    def _initialize_logger(self) -> None:
        bec_logger.configure(
            self.bootstrap_server,
            self._connector_cls,
            service_name=self._name,
            service_config=self._service_config.config["service_config"],
        )

    def _update_existing_services(self):
        service_keys = self.connector.keys(MessageEndpoints.service_status("*"))
        if not service_keys:
            return
        # service keys are in the form of: "internal/services/status/4b9d1af8-44ed-4f3a-8787-ef9f958f59b"
        services = []
        for service_key in service_keys:
            _, _, service_id = service_key.decode().rpartition("/")
            services.append(service_id)
        msgs = [
            self.connector.get(MessageEndpoints.service_status(service)) for service in services
        ]
        self._services_info = {msg.content["name"]: msg for msg in msgs if msg is not None}
        msgs = [self.connector.get(MessageEndpoints.metrics(service)) for service in services]
        self._services_metric = {msg.content["name"]: msg for msg in msgs if msg is not None}

    def _get_version_number(self):
        if self._version:
            return self._version

        bec_lib_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

        if os.path.exists(os.path.join(bec_lib_path, "pyproject.toml")):
            with open(os.path.join(bec_lib_path, "pyproject.toml"), "rb") as f:
                pyproject = tomli.load(f)
                self._version = pyproject["project"]["version"]
                return self._version
        self._version = importlib_version("bec-lib")
        return self._version

    def _update_service_info(self):
        while not self._service_info_event.is_set():
            logger.trace("Updating service info")
            try:
                self._send_service_status()
            except Exception:
                # exception is not explicitely specified,
                # because it depends on the underlying connector
                pass
            self._service_info_event.wait(timeout=3)

    def _send_service_status(self):
        version = self._get_version_number()
        self.connector.set_and_publish(
            topic=MessageEndpoints.service_status(self._service_id),
            msg=messages.StatusMessage(
                name=self._service_name,
                status=self.status,
                info={
                    "user": self._user,
                    "hostname": self._hostname,
                    "timestamp": time.time(),
                    "version": version,
                },
            ),
            expire=6,
        )

    @property
    def status(self) -> BECStatus:
        """get the current BECService status"""
        return self._status

    @status.setter
    def status(self, val: BECStatus):
        self._status = val
        self._send_service_status()

    def _start_update_service_info(self):
        self._service_info_thread = threading.Thread(
            target=self._update_service_info, daemon=True, name="update_service_info"
        )
        self._service_info_thread.start()

    def _start_metrics_emitter(self):
        self._metrics_emitter_thread = threading.Thread(
            target=self._get_metrics, daemon=True, name="metrics_emitter"
        )
        self._metrics_emitter_thread.start()

    def _get_metrics(self):
        proc = psutil.Process()
        while not self._metrics_emitter_event.is_set():
            res = proc.as_dict(
                attrs=[
                    "name",
                    "num_threads",
                    "pid",
                    "cpu_percent",
                    "memory_info",
                    "cmdline",
                    "cpu_times",
                    "create_time",
                    "memory_percent",
                ]
            )

            data = asdict(
                ServiceMetric(
                    process_name=res["name"],
                    username=self._user,
                    hostname=self._hostname,
                    cpu_percent=res["cpu_percent"],
                    cpu_times=res["cpu_times"].user,
                    cmdline=res["cmdline"],
                    num_threads=res["num_threads"],
                    pid=res["pid"],
                    memory_in_mb=res["memory_info"].rss / 1024 / 1024,
                    memory_used_percent=res["memory_percent"],
                    create_time=res["create_time"],
                )
            )
            msg = messages.ServiceMetricMessage(name=self._service_name, metrics=data)
            try:
                self.connector.set_and_publish(MessageEndpoints.metrics(self._service_id), msg)
            except Exception:
                # exception is not explicitely specified,
                # because it depends on the underlying connector
                pass
            self._metrics_emitter_event.wait(timeout=1)

    def set_global_var(self, name: str, val: Any) -> None:
        """Set a global variable through Redis

        Args:
            name (str): Name of the variable
            val (Any): Value of the variable

        """
        self.connector.set(MessageEndpoints.global_vars(name), messages.VariableMessage(value=val))

    def get_global_var(self, name: str) -> Any:
        """Get a global variable from Redis

        Args:
            name (str): Name of the variable

        Returns:
            Any: Value of the variable
        """
        msg = self.connector.get(MessageEndpoints.global_vars(name))
        if msg:
            return msg.content.get("value")
        return None

    def delete_global_var(self, name: str) -> None:
        """Delete a global variable from Redis

        Args:
            name (str): Name of the variable

        """
        self.connector.delete(MessageEndpoints.global_vars(name))

    def show_global_vars(self) -> str:
        """Get all available global variables"""
        # sadly, this cannot be a property as it causes side effects with IPython's tab completion
        available_keys = self.connector.keys(MessageEndpoints.global_vars("*"))

        def get_endpoint_from_topic(topic: str) -> str:
            return topic.decode().split(MessageEndpoints.global_vars("").endpoint)[-1]

        endpoints = [get_endpoint_from_topic(k) for k in available_keys]

        console = Console()
        table = Table(title="Global variables")
        table.add_column("Variable", justify="center")
        table.add_column("Content", justify="center")
        for endpoint in endpoints:
            var = str(self.get_global_var(endpoint))
            if len(var) > 40:
                var = var[0:20] + "..., " + var[-20:]
            table.add_row(endpoint, var)
        with console.capture() as capture:
            console.print(table)
        out = capture.get()
        logger.info(out)
        print(out)

    def shutdown(self):
        """shutdown the BECService"""
        try:
            self.connector.shutdown()
            self._service_info_event.set()
            if self._service_info_thread:
                self._service_info_thread.join()
            self._metrics_emitter_event.set()
            if self._metrics_emitter_thread:
                self._metrics_emitter_thread.join()
        except AttributeError:
            print("Failed to shutdown BECService.")

    @property
    def service_status(self):
        """get the status of active services"""
        self._update_existing_services()
        return self._services_info

    def wait_for_service(self, name, status=None):
        if status is None:
            status = BECStatus.RUNNING
        logger.info(f"Waiting for {name}.")
        while True:
            service_status_msg = self.service_status.get(name)
            if service_status_msg is not None:
                service_status = BECStatus(service_status_msg.content["status"])
                if service_status == status:
                    break
            time.sleep(0.05)
        logger.success(f"{name} is running.")

    def _wait_for_server(self):
        if not self.wait_for_server:
            return
        try:
            self.wait_for_service("ScanServer", BECStatus.RUNNING)
            self.wait_for_service("ScanBundler", BECStatus.RUNNING)
            self.wait_for_service("DeviceServer", BECStatus.RUNNING)
            self.wait_for_service("SciHub", BECStatus.RUNNING)
            logger.success("All BEC services are running.")
        except KeyboardInterrupt:
            logger.warning("KeyboardInterrupt received. Stopped waiting for BEC services.")


@dataclass
class ServiceMetric:
    """Container for keeping performance metrics."""

    pid: int
    cmdline: list
    process_name: str
    username: str
    hostname: str
    cpu_percent: float
    cpu_times: dict
    num_threads: int
    memory_in_mb: float
    memory_used_percent: float
    create_time: float



import enum
import inspect
import time
from typing import Generator, Protocol

from ophyd import DeviceStatus, Kind
from ophyd_devices.interfaces.protocols.bec_protocols import (
    BECDeviceProtocol,
    BECPositionerProtocol,
    BECSignalProtocol,
)

from bec_lib.devicemanager import DeviceContainer
from bec_lib.tests.utils import ConnectorMock

# pylint: disable=missing-function-docstring
# pylint: disable=protected-access


class DeviceMockType(enum.Enum):
    POSITIONER = "positioner"
    SIGNAL = "signal"


class DeviceObjectMock(BECDeviceProtocol):
    def __init__(self, name: str, kind=Kind.normal, dev_type=None, parent=None):
        self._name = name
        self._kind = kind if isinstance(kind, Kind) else getattr(Kind, kind)
        self._parent = parent
        self._dev_type = dev_type
        self._read_only = False
        self._enabled = True
        self._connected = True
        self._destroyed = False

    @property
    def destroyed(self):
        return self._destroyed

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, val):
        self._name = val

    @property
    def kind(self):
        return self._kind

    @kind.setter
    def kind(self, val):
        self._kind = val

    @property
    def parent(self):
        return self._parent

    @property
    def obj(self):
        return self

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, val: bool):
        self._enabled = val

    @property
    def read_only(self):
        return self._read_only

    @read_only.setter
    def read_only(self, val: bool):
        self._read_only = val

    @property
    def hints(self):
        if isinstance(self, PositionerMock):
            return {"fields": [self.name]}
        else:
            return {"fields": []}

    @property
    def root(self):
        return self if self.parent is None else self.parent

    @property
    def connected(self):
        return self._connected

    @connected.setter
    def connected(self, val):
        self._connected = val

    def destroy(self):
        self.connected = False
        self.enabled = False
        for _, obj in inspect.getmembers(self):
            if isinstance(obj, DeviceObjectMock) and obj.destroyed is False:
                obj.destroy()
        self._destroyed = True

    def trigger(self) -> None:
        pass


class MockSignal(DeviceObjectMock, BECSignalProtocol):

    def __init__(
        self, name: str, value: any = 0, kind: Kind = Kind.normal, parent=None, precision=None
    ):
        super().__init__(name=name, parent=parent, kind=kind)
        self._value = value
        self._config = {"deviceConfig": {"limits": [0, 0]}, "userParameter": None}
        self._metadata = dict(read_access=True, write_access=True, precision=precision)
        self._info = {
            "signals": {},
            "hints": self.hints,
            "write_access": self.write_access,
            "read_access": self.read_access,
        }

    def read(self):
        return {self.name: {"value": self._value, "timestamp": time.time()}}

    def read_configuration(self):
        return self.read()

    def get(self):
        return self._value

    def put(self, value: any, force: bool = False, timeout: float = None):
        self._value = value

    def set(self, value: any, timeout: float = None):
        self._value = value

    @property
    def metadata(self):
        return self._metadata.copy()

    @property
    def write_access(self):
        return self._metadata["write_access"]

    @property
    def read_access(self):
        return self._metadata["read_access"]

    @property
    def hints(self):
        if (~Kind.normal & Kind.hinted) & self.kind:
            return {"fields": [self.name]}
        else:
            return {"fields": []}

    @property
    def limits(self):
        return self._limits

    @limits.setter
    def limits(self, val):
        self._limits = val

    @property
    def low_limit(self):
        return self.limits[0]

    @property
    def high_limit(self):
        return self.limits[1]

    @property
    def precision(self):
        return self._metadata["precision"]

    def check_value(self, value):
        limits = self.limits
        if limits[0] == limits[1]:
            return
        if not limits[0] <= value <= limits[1]:
            raise ValueError(f"Value {value} is outside limits {limits}")


class PositionerMock(DeviceObjectMock, BECPositionerProtocol):

    def __init__(self, name: str, kind: Kind = Kind.normal, parent=None):
        super().__init__(name=name, parent=parent, kind=kind)
        # Name of readback signal is the same as the name of the positioner
        self.readback = MockSignal(name=self.name, kind=Kind.normal, value=0, precision=3)
        self.setpoint = MockSignal(name=f"{self.name}_setpoint", kind=Kind.normal, value=0)
        self.motor_is_moving = MockSignal(
            name=f"{self.name}_motor_is_moving", kind=Kind.normal, value=0
        )
        self.velocity = MockSignal(name=f"{self.name}_velocity", kind=Kind.config, value=0)
        self.acceleration = MockSignal(name=f"{self.name}_acceleration", kind=Kind.config, value=0)
        self._config = {"deviceConfig": {"limits": [-50, 50]}, "userParameter": None}
        self._read_attrs = ["readback", "setpoint", "motor_is_moving"]
        self._read_config_attrs = ["velocity", "acceleration"]
        self._info = {
            "signals": {
                name: {
                    "read_access": getattr(self, name).read_access,
                    "write_access": getattr(self, name).write_access,
                }
                for name in self._walk_components()
            },
            "hints": self.hints,
        }

    def _walk_components(self) -> Generator[str, None, None]:
        for name, obj in inspect.getmembers(self):
            if isinstance(obj, MockSignal):
                yield name

    def read(self):
        ret = {}
        for name in self._read_attrs:
            ret.update(getattr(self, name).read())
        return ret

    def read_configuration(self):
        ret = {}
        for name in self._read_config_attrs:
            ret.update(getattr(self, name).read())
        return ret

    def describe_configuration(self) -> dict:
        ret = {}
        for name in self._read_config_attrs:
            ret.update(
                {
                    name: {
                        "source": getattr(self, name).__class__.__name__,
                        "dtype": type(getattr(self, name)._value),
                        "shape": [],
                    }
                }
            )
        return ret

    def describe(self) -> dict:
        ret = {}
        for name in self._read_attrs:
            ret.update(
                {
                    name: {
                        "source": getattr(self, name).__class__.__name__,
                        "dtype": type(getattr(self, name)._value),
                        "shape": [],
                    }
                }
            )
        return ret

    @property
    def user_parameter(self):
        return self._config["userParameter"]

    @property
    def precision(self):
        return self.readback.precision

    @property
    def limits(self):
        return self._config["deviceConfig"]["limits"]

    @limits.setter
    def limits(self, val: tuple):
        self._config["deviceConfig"]["limits"] = val

    @property
    def low_limit(self):
        return self.limits[0]

    @property
    def high_limit(self):
        return self.limits[1]

    def check_value(self, value):
        limits = self.limits
        if limits[0] == limits[1]:
            return
        if not limits[0] <= value <= limits[1]:
            raise ValueError(f"Value {value} is outside limits {limits}")

    def move(self, position: float, wait=False, **kwargs) -> None:
        self.check_value(position)
        self.setpoint.put(position)
        self.readback.put(position)

    def set(self, position: float, **kwargs):
        self.move(position)


class DMMock:
    devices = DeviceContainer()
    connector = ConnectorMock()

    def add_device(self, name, value=None, dev_type: DeviceMockType = DeviceMockType.POSITIONER):
        if dev_type == DeviceMockType.POSITIONER:
            self.devices[name] = PositionerMock(name=name)
        elif dev_type == DeviceMockType.SIGNAL:
            self.devices[name] = MockSignal(name=name)
        else:
            raise ValueError(f"Unknown device type {dev_type}")
        if value is not None:
            self.devices[name].readback.put(value)



from __future__ import annotations

import threading
import time
from queue import Queue
from typing import TYPE_CHECKING

from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger

from .emitter import EmitterBase

logger = bec_logger.logger

if TYPE_CHECKING:
    from .scan_bundler import ScanBundler


class BECEmitter(EmitterBase):
    def __init__(self, scan_bundler: ScanBundler) -> None:
        super().__init__(scan_bundler.connector)
        self._send_buffer = Queue()
        self.scan_bundler = scan_bundler
        self._buffered_connector_thread = None
        self._buffered_publisher_stop_event = threading.Event()
        self._start_buffered_connector()

    def _start_buffered_connector(self):
        self._buffered_connector_thread = threading.Thread(
            target=self._buffered_publish, daemon=True, name="buffered_publisher"
        )
        self._buffered_connector_thread.start()

    def add_message(self, msg: messages.BECMessage, endpoint: str, public: str = None):
        self._send_buffer.put((msg, endpoint, public))

    def _buffered_publish(self):
        try:
            while not self._buffered_publisher_stop_event.is_set():
                self._publish_data()
        except Exception as e:
            logger.error(f"Error in buffered publisher: {e}")
            raise

    def _get_messages_from_buffer(self) -> list:
        msgs_to_send = []
        while not self._send_buffer.empty():
            msgs_to_send.append(self._send_buffer.get())
        return msgs_to_send

    def _publish_data(self) -> None:
        msgs_to_send = self._get_messages_from_buffer()

        if not msgs_to_send:
            time.sleep(0.1)
            return

        pipe = self.connector.pipeline()
        msgs = messages.BundleMessage()
        _, endpoint, _ = msgs_to_send[0]
        for msg, endpoint, public in msgs_to_send:
            msg_dump = msg
            msgs.append(msg_dump)
            if public:
                self.connector.set(public, msg_dump, pipe=pipe, expire=1800)
        self.connector.send(endpoint, msgs, pipe=pipe)
        pipe.execute()

    def on_scan_point_emit(self, scan_id: str, point_id: int):
        self._send_bec_scan_point(scan_id, point_id)

    def on_baseline_emit(self, scan_id: str):
        self._send_baseline(scan_id)

    def _send_bec_scan_point(self, scan_id: str, point_id: int) -> None:
        sb = self.scan_bundler

        info = sb.sync_storage[scan_id]["info"]
        msg = messages.ScanMessage(
            point_id=point_id,
            scan_id=scan_id,
            data=sb.sync_storage[scan_id][point_id],
            metadata={
                "scan_id": scan_id,
                "scan_type": info.get("scan_type"),
                "scan_report_devices": info.get("scan_report_devices"),
            },
        )
        self.add_message(
            msg,
            MessageEndpoints.scan_segment(),
            MessageEndpoints.public_scan_segment(scan_id=scan_id, point_id=point_id),
        )
        self._update_scan_progress(scan_id, point_id)

    def _update_scan_progress(self, scan_id: str, point_id: int, done=False) -> None:
        info = self.scan_bundler.sync_storage[scan_id]["info"]
        msg = messages.ProgressMessage(
            value=point_id + 1,
            max_value=info.get("num_points", point_id + 1),
            done=done,
            metadata={
                "scan_id": scan_id,
                "RID": info.get("RID", ""),
                "queue_id": info.get("queue_id", ""),
            },
        )
        self.scan_bundler.connector.set_and_publish(MessageEndpoints.scan_progress(), msg)

    def _send_baseline(self, scan_id: str) -> None:
        sb = self.scan_bundler

        msg = messages.ScanBaselineMessage(
            scan_id=scan_id,
            data=sb.sync_storage[scan_id]["baseline"],
            metadata=sb.sync_storage[scan_id]["info"],
        )
        pipe = sb.connector.pipeline()
        sb.connector.set(
            MessageEndpoints.public_scan_baseline(scan_id=scan_id), msg, expire=1800, pipe=pipe
        )
        sb.connector.set_and_publish(MessageEndpoints.scan_baseline(), msg, pipe=pipe)
        pipe.execute()

    def on_scan_status_update(self, status_msg: messages.ScanStatusMessage):
        if status_msg.status == "open":
            return
        num_points = status_msg.info.get("num_points", 0) - 1
        self._update_scan_progress(status_msg.scan_id, num_points, done=True)

    def shutdown(self):
        if self._buffered_connector_thread:
            self._buffered_publisher_stop_event.set()
            self._buffered_connector_thread.join()
            self._buffered_connector_thread = None
